from .base import Interface
from .decorators import endpoint
from ..definitions.types import ClientExtensions
from ..definitions.types import InstrumentName
from ..definitions.types import StopLossDetails
from ..definitions.types import TakeProfitDetails
from ..definitions.types import TradeID
from ..definitions.types import TradeSpecifier
from ..definitions.types import TradeStateFilter
from ..definitions.types import TrailingStopLossDetails
from ..endpoints.annotations import Count
from ..endpoints.annotations import Ids
from ..endpoints.annotations import Units
from ..endpoints.trade import *

__all__ = ['TradeInterface']


class TradeInterface(Interface):
    @endpoint(GETTrades)
    def list_trades(self, ids: Ids, state: TradeStateFilter, instrument: InstrumentName, count: Count,
                    before_id: TradeID):
        """
        Get a list of Trades for an Account

        Args:
            ids:
                List of Trade IDs to retrieve.
            state:
                The state to filter the requested Trades by.
            instrument:
                The instrument to filter the requested Trades by.
            count:
                The maximum number of Trades to return.
            before_id:
                The maximum Trade ID to return. If not provided the most recent
                Trades in the Account are returned.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETOpenTrades)
    def list_open_trades(self):
        """
        Get the list of open Trades for an Account

        Args:

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETTradeSpecifier)
    def get_trades(self, trade_specifier: TradeSpecifier):
        """
        Get the details of a specific Trade in an Account

        Args:
            trade_specifier:
                Specifier for the Trade

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(PUTTradeSpecifierClose)
    def close_trades(self, trade_specifier: TradeSpecifier, units: Units):
        """
        Close (partially or fully) a specific open Trade in an Account

        Args:
            trade_specifier:
                Specifier for the Trade
            units:
                Indication of how much of the Trade to close. Either the string
                "ALL" (indicating that all of the Trade should be closed), or a
                DecimalNumber representing the number of units of the open
                Trade to Close using a TradeClose MarketOrder. The units
                specified must always be positive, and the magnitude of the
                value cannot exceed the magnitude of the Trade's open units.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(PUTTradeSpecifierClientExtensions)
    def set_client_extensions_trade(self, trade_specifier: TradeSpecifier, client_extensions: ClientExtensions):
        """
        Update the Client Extensions for a Trade. Do not add, update, or delete
        the Client Extensions if your account is associated with MT4.

        Args:
            trade_specifier:
                Specifier for the Trade
            client_extensions:
                The Client Extensions to update the Trade with. Do not add,
                update, or delete the Client Extensions if your account is
                associated with MT4.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(PUTTradesSpecifierOrders)
    def set_dependent_orders_trade(self, trade_specifier: TradeSpecifier, take_profit: TakeProfitDetails,
                                   stop_loss: StopLossDetails, trailing_stop_loss: TrailingStopLossDetails):
        """
        Create, replace and cancel a Trade's dependent Orders (Take Profit,
        Stop Loss and Trailing Stop Loss) through the Trade itself

        Args:
            trade_specifier:
                Specifier for the Trade
            take_profit:
                The specification of the Take Profit to create/modify/cancel.
                If takeProfit is set to null, the Take Profit Order will be
                cancelled if it exists. If takeProfit is not provided, the
                existing Take Profit Order will not be modified. If a sub-
                field of takeProfit is not specified, that field will be set to
                a default value on create, and be inherited by the replacing
                order on modify.
            stop_loss:
                The specification of the Stop Loss to create/modify/cancel. If
                stopLoss is set to null, the Stop Loss Order will be cancelled
                if it exists. If stopLoss is not provided, the existing Stop
                Loss Order will not be modified. If a sub-field of stopLoss is
                not specified, that field will be set to a default value on
                create, and be inherited by the replacing order on modify.
            trailing_stop_loss:
                The specification of the Trailing Stop Loss to
                create/modify/cancel. If trailingStopLoss is set to null, the
                Trailing Stop Loss Order will be cancelled if it exists. If
                trailingStopLoss is not provided, the existing Trailing Stop
                Loss Order will not be modified. If a sub-field of
                trailingStopLoss is not specified, that field will be set to a
                default value on create, and be inherited by the replacing
                order on modify.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass



import pytest
from async_v20.definitions.base import Model
from async_v20.definitions.types import Account
from async_v20.definitions.helpers import flatten_dict

from ..data.json_data import GETAccountID_response

import ujson as json


@pytest.fixture
def account():
    result = Account(**GETAccountID_response['account'])
    yield result
    del result


def test_account_has_correct_methods(account):
    assert hasattr(account, 'json_dict')
    assert hasattr(account, 'data')
    assert hasattr(account, 'series')


@pytest.fixture
def test_kwargs():
    kwargs = {'type': 'LIST', 'value': 'TEST_VALUE'}
    yield kwargs
    del kwargs


@pytest.fixture
def test_class():
    class TestClass(Model):
        _dispatch = {'type': 'LIST'}

    test_cls = TestClass
    yield test_cls
    del test_cls


def test_base_dispatch_works_correctly():
    pass


def test_json_dict_returns_correct_data_structure(account):
    """Test the result is formatted correctly. There is a requirement for
    json_dict to be able to cast floats to strings, this is necessary when
    serializing objects to send to OANDA. Though when used internally is it more
    natural to leave floats as floats."""

    result = account.json_dict(float_to_string=True)
    # Test result is a dict
    assert type(result) == dict
    flattened_result = flatten_dict(result)
    # Test that all values have the correct data type
    for value in flattened_result:
        assert isinstance(value, (dict, str, int, list))

    result = account.json_dict(float_to_string=False)
    assert type(result) == dict
    flattened_result = flatten_dict(result)
    # Test that all values have the correct data type. Specifically that
    # all floats have not been casted to a string.
    for value in flattened_result:
        assert isinstance(value, (dict, float, str, int, list))
        if isinstance(value, str):
            with pytest.raises(ValueError):
                float(value)

def test_json_data(account):
    result = account.json_data()
    assert type(result) == str
    assert json.loads(result) == account.json_dict(float_to_string=True)

def test_data(account):
    result = account.data(float_to_string=True)
    for value in result:
        assert isinstance(value, (str, int, list))

    result = account.data(float_to_string=False)
    for value in result:
        print(value)
        assert isinstance(value, (float, str, int, list))
        if isinstance(value, str):
            with pytest.raises(ValueError):
                float(value)

def test_series(account):
    result = account.series()
    print(result)

    for value in result:
        assert isinstance(value, (float, str, int, list, type(None)))
        if isinstance(value, str):
            print(value)
            with pytest.raises(ValueError):
                float(value)



import asyncio

from async_v20 import OandaClient


async def account():
    async with OandaClient() as client:
        return await client.get_account_details()



loop = asyncio.get_event_loop()
response = loop.run_until_complete(account())

# pandas Series
print(response['account'].series())

# HTTP response state
print(response['account'])

# JSON data in python dictionary format
print(response['account'].json_dict())



import inspect

import pytest
from async_v20 import endpoints
from async_v20 import interface
from async_v20.client import OandaClient
from async_v20.definitions.types import Account
from async_v20.definitions.types import StopLossOrderRequest
from async_v20.definitions.types import AccountID
from async_v20.endpoints.annotations import Authorization
from async_v20.endpoints import POSTOrders
from async_v20.interface.helpers import _arguments
from async_v20.interface.helpers import _create_request_params
from async_v20.interface.helpers import create_url
from async_v20.interface.helpers import create_annotation_lookup
from async_v20.interface.helpers import create_body
from async_v20.interface.helpers import create_request_kwargs
from async_v20.interface.helpers import make_args_optional
from hypothesis.strategies import text, sampled_from

from .helpers import order_dict
from ..data.json_data import GETAccountID_response

client_attrs = [getattr(OandaClient, attr) for attr in dir(OandaClient)]
client_methods = list(filter(lambda x: hasattr(x, 'endpoint'), client_attrs))


@pytest.fixture
def stop_loss_order():
    order = StopLossOrderRequest(trade_id=1234, price=0.8)
    yield order
    del order


@pytest.fixture
def client():
    client = OandaClient(token='test_token', rest_host='rest_test', stream_host='stream_test')
    client.default_parameters = {}
    yield client
    del client


@pytest.mark.parametrize('method', client_methods)
def test_make_args_optional(method):
    """Ensure that all arguments passed to endpoint's are optional
    """
    result = make_args_optional(inspect.signature(method))

    def check_valid_param(param):
        if param.default != inspect._empty:
            return True
        else:
            print(param)

    assert all(map(check_valid_param, result.parameters.values()))


text_gen = text()
client_signatures = [inspect.signature(method) for method in client_methods]


def bound_args(sig):
    args = [text_gen.example() for _ in range(len(sig.parameters.keys()))]
    bound = sig.bind(*args)
    return sig, bound.arguments, args


annotation_lookup_arguments = [bound_args(sig) for sig in client_signatures]


@pytest.mark.asyncio
@pytest.mark.parametrize('signature, bound_arguments, args', annotation_lookup_arguments)
async def test_create_annotation_lookup(signature, bound_arguments, args):
    """Ensure that the annotation lookup dictionary is built correctly"""
    result = create_annotation_lookup(signature, bound_arguments)
    annotations = [param.annotation for param in signature.parameters.values()]
    correct = zip(annotations, args)
    assert all(map(lambda x: result[x[0]] == x[1], correct))


param_locations = ['header', 'path', 'query']
location = sampled_from(param_locations)
test_arguments_arguments = [(getattr(endpoints, cls), location.example()) for cls in endpoints.__all__]


@pytest.mark.parametrize('endpoint, param_location', test_arguments_arguments)
def test_arguments(endpoint, param_location):
    result = _arguments(endpoint, param_location)
    correct = list(filter(lambda x: x['located'] == param_location, endpoint.parameters))
    assert len(list(result)) == len(list(correct))


test_arguments_arguments = [(getattr(endpoints, cls), location.example(),) for cls in endpoints.__all__]


@pytest.mark.parametrize('interface_method', [method for cls in (getattr(interface, cls) for cls in interface.__all__)
                                              for method in cls.__dict__.values() if hasattr(method, 'endpoint')])
@pytest.mark.asyncio
async def test_create_request_params(client, interface_method):
    endpoint = interface_method.endpoint
    sig = interface_method.__signature__
    print(interface_method.__name__)
    args = tuple(range(len(sig.parameters)))
    bound = dict(sig.bind(*args).arguments)
    arguments = create_annotation_lookup(sig, bound)
    total_params = []
    print(endpoint.request_schema)
    for location in param_locations:
        print('Endpoint: ', endpoint)
        print('Arguments: ', arguments)
        print('Location: ', location)
        result = _create_request_params(client, endpoint, arguments, location)
        print('Possible Arguments', list(_arguments(endpoint, arguments)))
        print(location, ': ', result)
        total_params.extend(result)

    assert len(total_params) == len(arguments) - len(list(endpoint.request_schema)) -1  #-1 removes 'self'

@pytest.mark.parametrize('endpoint', [getattr(endpoints, cls) for cls in endpoints.__all__])
def test_create_url(client, endpoint):
    path = endpoint.path.path
    arguments = [value for value in path if not isinstance(value, str)]
    values = list(map(lambda x: str(x), range(len(arguments))))
    arguments = dict(zip(arguments, values))
    url = create_url(client, endpoint, arguments)
    path = url.path
    for value in values:
        assert value in path
        path = path[path.index(value):]

@pytest.mark.parametrize('interface_method', [method for cls in (getattr(interface, cls) for cls in interface.__all__)
                                              for method in cls.__dict__.values() if hasattr(method, 'endpoint')])
def test_create_request_kwargs(client, interface_method):
    client.default_parameters.update({AccountID: 'TEST_ID',
                                      Authorization: 'TEST_AUTH'})
    args = list(map(lambda x: str(x), (range(len(interface_method.__signature__.parameters)))))[1:]
    print(interface_method.__name__)
    request_kwargs = create_request_kwargs(client,
                                           interface_method.endpoint,
                                           interface_method.__signature__,
                                           *args)
    print("('",request_kwargs['method'], "', '", request_kwargs['url'].path, "')")

    assert 'method' in request_kwargs
    assert 'url' in request_kwargs
    assert 'headers' in request_kwargs
    assert 'params' in request_kwargs
    assert 'json' in request_kwargs

    assert [request_kwargs['method']] in [['POST'], ['GET'], ['PUT'], ['PATCH'], ['DELETE']]
    assert 'Authorization' in request_kwargs['headers']


@pytest.mark.asyncio
async def test_request_body_is_constructed_correctly(stop_loss_order):
    result = create_body(POSTOrders.request_schema,
                         {'irrelevant': stop_loss_order, 'test': Account(), 'arg': 'random_string'})
    print(result)
    assert result == {'order': {'tradeID': 1234, 'price': '0.8', 'type': 'STOP_LOSS', 'timeInForce': 'GTC',
                                'triggerCondition': 'DEFAULT'}}


@pytest.mark.asyncio
async def test_objects_can_be_converted_between_Model_object_and_json():
    account = Account(**GETAccountID_response['account'])
    response_json_account = GETAccountID_response['account']
    account_to_json = account.json_dict()

    response_json_account = order_dict(response_json_account)
    account_to_json = order_dict(account_to_json)
    print('SERVER DATA')
    print(response_json_account)
    print('ASYNC_20 DATA')
    print(account_to_json)
    assert response_json_account == account_to_json



import asyncio

from async_v20 import OandaClient

client = OandaClient()

loop = asyncio.get_event_loop()
response = loop.run_until_complete(client.initialize())

# Write your code here

client.close()


from .annotations import *
from .base import EndPoint, Path
from ..definitions.descriptors import *
from ..definitions.metaclass import Array
from ..definitions.types import *

__all__ = ['GETTransactions', 'GETTransactionID', 'GETIDrange', 'GETSinceID', 'GETTransactionsStream']


class GETTransactions(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/transactions')

    # description of endpoint
    description = 'Get a list of Transactions pages that satisfy a time-based Transaction query.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'from', 'located': 'query', 'type': FromTime, 'description': 'DateTime'},
        {'name': 'to', 'located': 'query', 'type': ToTime, 'description': 'DateTime'},
        {'name': 'pageSize', 'located': 'query', 'type': PageSize, 'description': 'int'},
        {'name': 'type', 'located': 'query', 'type': Type,
         'description': 'List of TransactionFilter (csv)'},
    ]

    # valid responses
    responses = {
        200: {'from': DateTime, 'to': DateTime, 'pageSize': int, 'type': Array(TransactionFilter), 'count': int,
              'pages': Array(str), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (400, 401, 403, 404, 405, 416)


class GETTransactionID(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/transactions/', TransactionID)

    # description of endpoint
    description = 'Get the details of a single Account Transaction.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'transactionID', 'located': 'path', 'type': TransactionID, 'description': 'TransactionID'},
    ]

    # valid responses
    responses = {200: {'transaction': Transaction, 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class GETIDrange(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/transactions/idrange')

    # description of endpoint
    description = 'Get a range of Transactions for an Account based on the Transaction IDs.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'from', 'located': 'query', 'type': FromTransactionID, 'description': 'TransactionID'},
        {'name': 'to', 'located': 'query', 'type': ToTransactionID, 'description': 'TransactionID'},
        {'name': 'type', 'located': 'query', 'type': Type,
         'description': 'List of TransactionFilter (csv)'},
    ]

    # valid responses
    responses = {200: {'transactions': Array(Transaction), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (400, 401, 404, 405, 416)


class GETSinceID(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/transactions/sinceid')

    # description of endpoint
    description = 'Get a range of Transactions for an Account starting at ' \
                  '(but not including) a provided Transaction ID.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'id', 'located': 'query', 'type': TransactionID, 'description': 'TransactionID'},
    ]

    # valid responses
    responses = {200: {'transactions': Array(Transaction), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (400, 401, 404, 405, 416)


class GETTransactionsStream(EndPoint):
    # host to use for this endpoint
    host = 'STREAM'

    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/transactions/stream')

    # description of endpoint
    description = 'Get a stream of Transactions for an Account starting from when the request is made.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {200: {'transaction': Transaction, 'transactionHeartbeat': TransactionHeartbeat}}

    # error msgs'
    error = (400, 401, 404, 405)



import asyncio

from async_v20 import OandaClient

client = OandaClient()


async def poll_account(poll_interval=6, client=client):
    while True:
        result = await client.account_changes()
        print(result)
        await asyncio.sleep(poll_interval)


async def stream(instruments, client=client):
    async for data in await client.stream_pricing(instruments):
        price = data.get('PRICE', None)
        if price:
            print(price.series())


loop = asyncio.get_event_loop()
loop.run_until_complete(
    asyncio.gather(poll_account(), stream('EUR_USD'))
    )
client.close()



from functools import partial
from inspect import Signature, _empty

from ..helpers import sleep


def make_args_optional(signature):
    sig = Signature([param.replace(default=None)
                     if param.default == _empty
                     else param
                     for param in signature.parameters.values()])
    return sig


def create_annotation_lookup(signature, bound_arguments):
    """Combine the signatures annotations with bound arguments to create a lookup dict
    for subsequent functions to identify arguments they need to use"""
    annotations_lookup = {param.name: param.annotation for param in signature.parameters.values()}
    return {annotations_lookup[name]: value for name, value in bound_arguments.items()}


def _arguments(endpoint, param_location):
    return ((parameter['name'], parameter['type']) for parameter in endpoint.parameters if
            parameter['located'] == param_location)


def _create_request_params(self, endpoint, arguments: dict, param_location: str):
    possible_arguments = _arguments(endpoint, param_location)

    def lookup(typ):
        result = None
        try:
            result = arguments[typ]
        except KeyError:
            try:
                result = self.default_parameters[typ]
            except TypeError:
                # TODO create exception module
                raise Exception('No default parameters provided')
            except KeyError:
                print(f"WARNING: missing {typ.__name__} in {param_location}")
                pass  # TODO: This Should raise a warning that not all header parameters were created
        return result

    parameters = ((name, lookup(typ)) for name, typ in possible_arguments)
    return {name: value for name, value in parameters if value is not None}


def create_url(self, endpoint, arguments):
    endpoint_path = endpoint.path(arguments, default=self.default_parameters)
    host = self.hosts[endpoint.host]
    return host(path=endpoint_path)


def create_body(request_schema, arguments):
    # Reverse the request schema to allow for lookups
    lookup = {value: key for key, value in request_schema.items()}

    def dumps():
        """Iterate over the arguments returning json_dicts of matching objects"""
        for argument in arguments.values():
            try:
                key = lookup.get(argument._derived, None)
            except AttributeError:
                continue
            else:
                if key:
                    yield (key, argument.json_dict())

    return dict([json_data for json_data in dumps()])


header_params = partial(_create_request_params, param_location='header')

query_params = partial(_create_request_params, param_location='query')


def create_request_kwargs(self, endpoint, sig, *args, **kwargs):
    """Format arguments to be passed to an aiohttp request"""
    arguments = sig.bind(self, *args, **kwargs).arguments
    arguments = create_annotation_lookup(sig, arguments)

    json = create_body(endpoint.request_schema, arguments)

    headers = header_params(self, endpoint, arguments)
    url = create_url(self, endpoint, arguments)
    parameters = query_params(self, endpoint, arguments)

    request_kwargs = {
        'method': endpoint.method,
        'url': url,
        'headers': headers,
        'params': parameters,
        'json': json,
    }

    if endpoint.host == 'STREAM':
        request_kwargs.update({'timeout': 0})

    return request_kwargs



from .base import Descriptor

__all__ = ['PriceStatus', 'PriceValue']


class PriceStatus(Descriptor):
    """The status of the Price.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'tradeable': 'The Instrument’s price is tradeable.',
        'non-tradeable': 'The Instrument’s price is not tradeable.',
        'invalid': 'The Instrument of the price is invalid or there is no valid Price for the Instrument.'
    }


class PriceValue(Descriptor):
    """The string representation of a Price for an Instrument.
    """

    # Type checking
    typ = float

    # Correct syntax of value
    format_syntax = 'A decimal number encodes as a string. The amount of precision ' \
                    'provided depends on the Price’s Instrument.'

    def __set__(self, instance, value):
        value = super().type_check(value)
        super().__set__(instance, round(value, 5))


from .base import Descriptor

__all__ = ['ClientComment', 'ClientID', 'ClientTag', 'FundingReason', 'LimitOrderReason', 'MarketIfTouchedOrderReason',
           'MarketOrderMarginCloseoutReason', 'MarketOrderReason', 'OrderCancelReason', 'OrderFillReason', 'RequestID',
           'StopLossOrderReason', 'StopOrderReason', 'TakeProfitOrderReason', 'TrailingStopLossOrderReason',
           'TransactionFilter', 'TransactionID', 'TransactionRejectReason', 'TransactionType']


class ClientComment(Descriptor):
    """A client-provided comment that can contain any data and may be assigned to their Orders or
    Trades. Comments are typically used to provide extra context or meaning to an Order or Trade.
    """

    # Type checking
    typ = str

    # Example of correct format
    example = 'This is a client comment'


class ClientID(Descriptor):
    """A client-provided identifier, used by clients to refer to their
    Orders or Trades with an identifier that they have provided.
    """

    # Type checking
    typ = str

    # Example of correct format
    example = 'my_order_id'


class ClientTag(Descriptor):
    """A client-provided tag that can contain any data and may be assigned to their
    Orders or Trades. Tags are typically used to associate groups of Trades and/or Orders together.
    """

    # Type checking
    typ = str

    # Example of correct format
    example = 'client_tag_1'


class FundingReason(Descriptor):
    """The reason that an Account is being funded.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CLIENT_FUNDING': 'The client has initiated a funds transfer',
        'ACCOUNT_TRANSFER': 'Funds are being transfered between two Accounts.',
        'DIVISION_MIGRATION': 'Funds are being transfered as part of a Division migration',
        'SITE_MIGRATION': 'Funds are being transfered as part of a Site migration',
        'ADJUSTMENT': 'Funds are being transfered as part of an Account adjustment'
    }


class LimitOrderReason(Descriptor):
    """The reason that the Limit Order was initiated
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CLIENT_ORDER': 'The Limit Order was initiated at the request of a client',
        'REPLACEMENT': 'The Limit Order was initiated as a replacement for an existing Order'
    }


class MarketIfTouchedOrderReason(Descriptor):
    """The reason that the Market-if-touched Order was initiated
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CLIENT_ORDER': 'The Market-if-touched Order was initiated at the request of a client',
        'REPLACEMENT': 'The Market-if-touched Order was initiated as a replacement for an existing Order'
    }


class MarketOrderMarginCloseoutReason(Descriptor):
    """The reason that the Market Order was created to perform a margin closeout
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'MARGIN_CHECK_VIOLATION': 'Trade closures resulted from violating OANDA’s margin policy',
        'REGULATORY_MARGIN_CALL_VIOLATION': 'Trade closures came from a margin closeout event resulting from '
                                            'regulatory conditions placed on the Account’s margin call'
    }


class MarketOrderReason(Descriptor):
    """The reason that the Market Order was created
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CLIENT_ORDER': 'The Market Order was created at the request of a client',
        'TRADE_CLOSE': 'The Market Order was created to close a Trade at the request of a client',
        'POSITION_CLOSEOUT': 'The Market Order was created to close a Position at the request of a client',
        'MARGIN_CLOSEOUT': 'The Market Order was created as part of a Margin Closeout',
        'DELAYED_TRADE_CLOSE': 'The Market Order was created to close a trade marked for delayed closure'
    }


class OrderCancelReason(Descriptor):
    """The reason that an Order was cancelled.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'INTERNAL_SERVER_ERROR': 'The Order was cancelled because at the time of filling, '
                                 'an unexpected internal server error occurred.',
        'ACCOUNT_LOCKED': 'The Order was cancelled because at the time of filling the account was locked.',
        'ACCOUNT_NEW_POSITIONS_LOCKED': 'The order was to be filled, however the account is configured to not allow new'
                                        'positions to be created.',
        'ACCOUNT_ORDER_CREATION_LOCKED': 'Filling the Order wasn’t possible because it required the creation of a '
                                         'dependent Order and the Account is locked for Order creation.',
        'ACCOUNT_ORDER_FILL_LOCKED': 'Filling the Order was not possible because the Account is locked for filling '
                                     'Orders.',
        'CLIENT_REQUEST': 'The Order was cancelled explicitly at the request of the client.',
        'MIGRATION': 'The Order cancelled because it is being migrated to another account.',
        'MARKET_HALTED': 'Filling the Order wasn’t possible because the Order’s instrument was halted.',
        'LINKED_TRADE_CLOSED': 'The Order is linked to an open Trade that was closed.',
        'TIME_IN_FORCE_EXPIRED': 'The time in force specified for this order has passed.',
        'INSUFFICIENT_MARGIN': 'Filling the Order wasn’t possible because the Account had insufficient margin.',
        'FIFO_VIOLATION': 'Filling the Order would have resulted in a a FIFO violation.',
        'BOUNDS_VIOLATION': 'Filling the Order would have violated the Order’s price bound.',
        'CLIENT_REQUEST_REPLACED': 'The Order was cancelled for replacement at the request of the client.',
        'INSUFFICIENT_LIQUIDITY': 'Filling the Order wasn’t possible because enough liquidity available.',
        'TAKE_PROFIT_ON_FILL_GTD_TIMESTAMP_IN_PAST': 'Filling the Order would have resulted in the creation of a '
                                                     'Take Profit Order with a GTD time in the past.',
        'TAKE_PROFIT_ON_FILL_LOSS': 'Filling the Order would result in the creation of a Take Profit Order that '
                                    'would have been filled immediately, closing the new Trade at a loss.',
        'LOSING_TAKE_PROFIT': 'Filling the Order would result in the creation of a Take Profit Loss Order that would '
                              'close the new Trade at a loss when filled.',
        'STOP_LOSS_ON_FILL_GTD_TIMESTAMP_IN_PAST': 'Filling the Order would have resulted in the creation of a Stop '
                                                   'Loss Order with a GTD time in the past.',
        'STOP_LOSS_ON_FILL_LOSS': 'Filling the Order would result in the creation of a Stop Loss Order that would have '
                                  'been filled immediately, closing the new Trade at a loss.',
        'TRAILING_STOP_LOSS_ON_FILL_GTD_TIMESTAMP_IN_PAST': 'Filling the Order would have resulted in the creation of '
                                                            'a Trailing Stop Loss Order with a GTD time in the past.',
        'CLIENT_TRADE_ID_ALREADY_EXISTS': 'Filling the Order would result in the creation of a new Open Trade with a '
                                          'client Trade ID already in use.',
        'POSITION_CLOSEOUT_FAILED': 'Closing out a position wasn’t fully possible.',
        'OPEN_TRADES_ALLOWED_EXCEEDED': 'Filling the Order would cause the maximum open trades allowed for the Account '
                                        'to be exceeded.',
        'PENDING_ORDERS_ALLOWED_EXCEEDED': 'Filling the Order would have resulted in exceeding the number of pending '
                                           'Orders allowed for the Account.',
        'TAKE_PROFIT_ON_FILL_CLIENT_ORDER_ID_ALREADY_EXISTS': 'Filling the Order would have resulted in the creation '
                                                              'of a Take Profit Order with a client Order ID that is '
                                                              'already in use.',
        'STOP_LOSS_ON_FILL_CLIENT_ORDER_ID_ALREADY_EXISTS': 'Filling the Order would have resulted in the creation of '
                                                            'a Stop Loss Order with a client Order ID that is already '
                                                            'in use.',
        'TRAILING_STOP_LOSS_ON_FILL_CLIENT_ORDER_ID_ALREADY_EXISTS': 'Filling the Order would have resulted in the '
                                                                     'creation of a Trailing Stop Loss Order with a '
                                                                     'client Order ID that is already in use.',
        'POSITION_SIZE_EXCEEDED': 'Filling the Order would have resulted in the Account’s maximum position size limit '
                                  'being exceeded for the Order’s instrument.'
    }


class OrderFillReason(Descriptor):
    """The reason that an Order was filled
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'LIMIT_ORDER': 'The Order filled was a Limit Order',
        'STOP_ORDER': 'The Order filled was a Stop Order',
        'MARKET_IF_TOUCHED_ORDER': 'The Order filled was a Market-if-touched Order',
        'TAKE_PROFIT_ORDER': 'The Order filled was a Take Profit Order',
        'STOP_LOSS_ORDER': 'The Order filled was a Stop Loss Order',
        'TRAILING_STOP_LOSS_ORDER': 'The Order filled was a Trailing Stop Loss Order',
        'MARKET_ORDER': 'The Order filled was a Market Order',
        'MARKET_ORDER_TRADE_CLOSE': 'The Order filled was a Market Order used to explicitly close a Trade',
        'MARKET_ORDER_POSITION_CLOSEOUT': 'The Order filled was a Market Order used to explicitly close a Position',
        'MARKET_ORDER_MARGIN_CLOSEOUT': 'The Order filled was a Market Order used for a Margin Closeout',
        'MARKET_ORDER_DELAYED_TRADE_CLOSE': 'The Order filled was a Market Order used for a delayed Trade close'
    }


class RequestID(Descriptor):
    """The request identifier.
    """

    # Type checking
    typ = str


class StopLossOrderReason(Descriptor):
    """The reason that the Stop Loss Order was initiated
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CLIENT_ORDER': 'The Stop Loss Order was initiated at the request of a client',
        'REPLACEMENT': 'The Stop Loss Order was initiated as a replacement for an existing Order',
        'ON_FILL': 'The Stop Loss Order was initiated automatically when an Order was filled that opened a new Trade '
                   'requiring a Stop Loss Order.'
    }


class StopOrderReason(Descriptor):
    """The reason that the Stop Order was initiated
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CLIENT_ORDER': 'The Stop Order was initiated at the request of a client',
        'REPLACEMENT': 'The Stop Order was initiated as a replacement for an existing Order'
    }


class TakeProfitOrderReason(Descriptor):
    """The reason that the Take Profit Order was initiated
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CLIENT_ORDER': 'The Take Profit Order was initiated at the request of a client',
        'REPLACEMENT': 'The Take Profit Order was initiated as a replacement for an existing Order',
        'ON_FILL': 'The Take Profit Order was initiated automatically when an Order was filled that opened a new Trade '
                   'requiring a Take Profit Order.'
    }


class TrailingStopLossOrderReason(Descriptor):
    """The reason that the Trailing Stop Loss Order was initiated
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CLIENT_ORDER': 'The Trailing Stop Loss Order was initiated at the request of a client',
        'REPLACEMENT': 'The Trailing Stop Loss Order was initiated as a replacement for an existing Order',
        'ON_FILL': 'The Trailing Stop Loss Order was initiated automatically when an Order was filled that opened a '
                   'new Trade requiring a Trailing Stop Loss Order.'}


class TransactionFilter(Descriptor):
    """A filter that can be used when fetching Transactions
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'ORDER': 'Order-related Transactions. These are the Transactions that create, cancel, fill or trigger Orders',
        'FUNDING': 'Funding-related Transactions',
        'ADMIN': 'Administrative Transactions',
        'CREATE': 'Account Create Transaction',
        'CLOSE': 'Account Close Transaction',
        'REOPEN': 'Account Reopen Transaction',
        'CLIENT_CONFIGURE': 'Client Configuration Transaction',
        'CLIENT_CONFIGURE_REJECT': 'Client Configuration Reject Transaction',
        'TRANSFER_FUNDS': 'Transfer Funds Transaction',
        'TRANSFER_FUNDS_REJECT': 'Transfer Funds Reject Transaction',
        'MARKET_ORDER': 'Market Order Transaction',
        'MARKET_ORDER_REJECT': 'Market Order Reject Transaction',
        'LIMIT_ORDER': 'Limit Order Transaction',
        'LIMIT_ORDER_REJECT': 'Limit Order Reject Transaction',
        'STOP_ORDER': 'Stop Order Transaction',
        'STOP_ORDER_REJECT': 'Stop Order Reject Transaction',
        'MARKET_IF_TOUCHED_ORDER': 'Market if Touched Order Transaction',
        'MARKET_IF_TOUCHED_ORDER_REJECT': 'Market if Touched Order Reject Transaction',
        'TAKE_PROFIT_ORDER': 'Take Profit Order Transaction',
        'TAKE_PROFIT_ORDER_REJECT': 'Take Profit Order Reject Transaction',
        'STOP_LOSS_ORDER': 'Stop Loss Order Transaction',
        'STOP_LOSS_ORDER_REJECT': 'Stop Loss Order Reject Transaction',
        'TRAILING_STOP_LOSS_ORDER': 'Trailing Stop Loss Order Transaction',
        'TRAILING_STOP_LOSS_ORDER_REJECT': 'Trailing Stop Loss Order Reject Transaction',
        'ONE_CANCELS_ALL_ORDER': 'One Cancels All Order Transaction',
        'ONE_CANCELS_ALL_ORDER_REJECT': 'One Cancels All Order Reject Transaction',
        'ONE_CANCELS_ALL_ORDER_TRIGGERED': 'One Cancels All Order Trigger Transaction',
        'ORDER_FILL': 'Order Fill Transaction',
        'ORDER_CANCEL': 'Order Cancel Transaction',
        'ORDER_CANCEL_REJECT': 'Order Cancel Reject Transaction',
        'ORDER_CLIENT_EXTENSIONS_MODIFY': 'Order Client Extensions Modify Transaction',
        'ORDER_CLIENT_EXTENSIONS_MODIFY_REJECT': 'Order Client Extensions Modify Reject Transaction',
        'TRADE_CLIENT_EXTENSIONS_MODIFY': 'Trade Client Extensions Modify Transaction',
        'TRADE_CLIENT_EXTENSIONS_MODIFY_REJECT': 'Trade Client Extensions Modify Reject Transaction',
        'MARGIN_CALL_ENTER': 'Margin Call Enter Transaction',
        'MARGIN_CALL_EXTEND': 'Margin Call Extend Transaction',
        'MARGIN_CALL_EXIT': 'Margin Call Exit Transaction',
        'DELAYED_TRADE_CLOSURE': 'Delayed Trade Closure Transaction',
        'DAILY_FINANCING': 'Daily Financing Transaction',
        'RESET_RESETTABLE_PL': 'Reset Resettable PL Transaction'
    }


class TransactionID(Descriptor):
    """The unique Transaction identifier within each Account.
    """

    # Type checking
    typ = str

    # Correct syntax of value
    format_syntax = 'String representation of the numerical OANDA-assigned TransactionID'
    # Example of correct format
    # example = '1523'


class TransactionRejectReason(Descriptor):
    """The reason that a Transaction was rejected.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'INTERNAL_SERVER_ERROR': 'An unexpected internal server error has occurred',
        'INSTRUMENT_PRICE_UNKNOWN': 'The system was unable to determine the current price for the Order’s instrument',
        'ACCOUNT_NOT_ACTIVE': 'The Account is not active',
        'ACCOUNT_LOCKED': 'The Account is locked',
        'ACCOUNT_ORDER_CREATION_LOCKED': 'The Account is locked for Order creation',
        'ACCOUNT_CONFIGURATION_LOCKED': 'The Account is locked for configuration',
        'ACCOUNT_DEPOSIT_LOCKED': 'The Account is locked for deposits',
        'ACCOUNT_WITHDRAWAL_LOCKED': 'The Account is locked for withdrawals',
        'ACCOUNT_ORDER_CANCEL_LOCKED': 'The Account is locked for Order cancellation',
        'INSTRUMENT_NOT_TRADEABLE': 'The instrument specified is not tradeable by the Account',
        'PENDING_ORDERS_ALLOWED_EXCEEDED': 'Creating the Order would result in the maximum number of allowed pending '
                                           'Orders being exceeded',
        'ORDER_ID_UNSPECIFIED': 'Neither the Order ID nor client Order ID are specified',
        'ORDER_DOESNT_EXIST': 'The Order specified does not exist',
        'ORDER_IDENTIFIER_INCONSISTENCY': 'The Order ID and client Order ID specified do not identify the same Order',
        'TRADE_ID_UNSPECIFIED': 'Neither the Trade ID nor client Trade ID are specified',
        'TRADE_DOESNT_EXIST': 'The Trade specified does not exist',
        'TRADE_IDENTIFIER_INCONSISTENCY': 'The Trade ID and client Trade ID specified do not identify the same Trade',
        'INSTRUMENT_MISSING': 'Order instrument has not been specified',
        'INSTRUMENT_UNKNOWN': 'The instrument specified is unknown',
        'UNITS_MISSING': 'Order units have not been not specified',
        'UNITS_INVALID': 'Order units specified are invalid',
        'UNITS_PRECISION_EXCEEDED': 'The units specified contain more precision than is allowed for the Order’s '
                                    'instrument',
        'UNITS_LIMIT_EXCEEDED': 'The units specified exceeds the maximum number of units allowed',
        'UNITS_MIMIMUM_NOT_MET': 'The units specified is less than the minimum number of units required',
        'PRICE_MISSING': 'The price has not been specified',
        'PRICE_INVALID': 'The price specifed is invalid',
        'PRICE_PRECISION_EXCEEDED': 'The price specified contains more precision than is allowed for the instrument',
        'PRICE_DISTANCE_MISSING': 'The price distance has not been specified',
        'PRICE_DISTANCE_INVALID': 'The price distance specifed is invalid',
        'PRICE_DISTANCE_PRECISION_EXCEEDED': 'The price distance specified contains more precision than is allowed for '
                                             'the instrument',
        'PRICE_DISTANCE_MAXIMUM_EXCEEDED': 'The price distance exceeds that maximum allowed amount',
        'PRICE_DISTANCE_MINIMUM_NOT_MET': 'The price distance does not meet the minimum allowed amount',
        'TIME_IN_FORCE_MISSING': 'The TimeInForce field has not been specified',
        'TIME_IN_FORCE_INVALID': 'The TimeInForce specified is invalid',
        'TIME_IN_FORCE_GTD_TIMESTAMP_MISSING': 'The TimeInForce is GTD but no GTD timestamp is provided',
        'TIME_IN_FORCE_GTD_TIMESTAMP_IN_PAST': 'The TimeInForce is GTD but the GTD timestamp is in the past',
        'PRICE_BOUND_INVALID': 'The price bound specified is invalid',
        'PRICE_BOUND_PRECISION_EXCEEDED': 'The price bound specified contains more precision than is allowed for the '
                                          'Order’s instrument',
        'ORDERS_ON_FILL_DUPLICATE_CLIENT_ORDER_IDS': 'Multiple Orders on fill share the same client Order ID',
        'TRADE_ON_FILL_CLIENT_EXTENSIONS_NOT_SUPPORTED': 'The Order does not support Trade on fill client extensions '
                                                         'because it cannot create a new Trade',
        'CLIENT_ORDER_ID_INVALID': 'The client Order ID specified is invalid',
        'CLIENT_ORDER_ID_ALREADY_EXISTS': 'The client Order ID specified is already assigned to another pending Order',
        'CLIENT_ORDER_TAG_INVALID': 'The client Order tag specified is invalid',
        'CLIENT_ORDER_COMMENT_INVALID': 'The client Order comment specified is invalid',
        'CLIENT_TRADE_ID_INVALID': 'The client Trade ID specified is invalid',
        'CLIENT_TRADE_ID_ALREADY_EXISTS': 'The client Trade ID specifed is already assigned to another open Trade',
        'CLIENT_TRADE_TAG_INVALID': 'The client Trade tag specified is invalid',
        'CLIENT_TRADE_COMMENT_INVALID': 'The client Trade comment is invalid',
        'ORDER_FILL_POSITION_ACTION_MISSING': 'The OrderFillPositionAction field has not been specified',
        'ORDER_FILL_POSITION_ACTION_INVALID': 'The OrderFillPositionAction specified is invalid',
        'TRIGGER_CONDITION_MISSING': 'The TriggerCondition field has not been specified',
        'TRIGGER_CONDITION_INVALID': 'The TriggerCondition specified is invalid',
        'ORDER_PARTIAL_FILL_OPTION_MISSING': 'The OrderFillPositionAction field has not been specified',
        'ORDER_PARTIAL_FILL_OPTION_INVALID': 'The OrderFillPositionAction specified is invalid.',
        'INVALID_REISSUE_IMMEDIATE_PARTIAL_FILL': 'When attempting to reissue an order (currently only a '
                                                  'MarketIfTouched) that was immediately partially filled, it is not '
                                                  'possible to create a correct pending Order.',
        'TAKE_PROFIT_ORDER_ALREADY_EXISTS': 'A Take Profit Order for the specified Trade already exists',
        'TAKE_PROFIT_ON_FILL_PRICE_MISSING': 'The Take Profit on fill specified does not provide a price',
        'TAKE_PROFIT_ON_FILL_PRICE_INVALID': 'The Take Profit on fill specified contains an invalid price',
        'TAKE_PROFIT_ON_FILL_PRICE_PRECISION_EXCEEDED': 'The Take Profit on fill specified contains a price with more '
                                                        'precision than is allowed by the Order’s instrument',
        'TAKE_PROFIT_ON_FILL_TIME_IN_FORCE_MISSING': 'The Take Profit on fill specified does not provide a TimeInForce',
        'TAKE_PROFIT_ON_FILL_TIME_IN_FORCE_INVALID': 'The Take Profit on fill specifies an invalid TimeInForce',
        'TAKE_PROFIT_ON_FILL_GTD_TIMESTAMP_MISSING': 'The Take Profit on fill specifies a GTD TimeInForce but does not '
                                                     'provide a GTD timestamp',
        'TAKE_PROFIT_ON_FILL_GTD_TIMESTAMP_IN_PAST': 'The Take Profit on fill specifies a GTD timestamp that is in the '
                                                     'past',
        'TAKE_PROFIT_ON_FILL_CLIENT_ORDER_ID_INVALID': 'The Take Profit on fill client Order ID specified is invalid',
        'TAKE_PROFIT_ON_FILL_CLIENT_ORDER_TAG_INVALID': 'The Take Profit on fill client Order tag specified is invalid',
        'TAKE_PROFIT_ON_FILL_CLIENT_ORDER_COMMENT_INVALID': 'The Take Profit on fill client Order comment specified is'
                                                            ' invalid',
        'TAKE_PROFIT_ON_FILL_TRIGGER_CONDITION_MISSING': 'The Take Profit on fill specified does not provide a '
                                                         'TriggerCondition',
        'TAKE_PROFIT_ON_FILL_TRIGGER_CONDITION_INVALID': 'The Take Profit on fill specifies an invalid '
                                                         'TriggerCondition',
        'STOP_LOSS_ORDER_ALREADY_EXISTS': 'A Stop Loss Order for the specified Trade already exists',
        'STOP_LOSS_ON_FILL_PRICE_MISSING': 'The Stop Loss on fill specified does not provide a price',
        'STOP_LOSS_ON_FILL_PRICE_INVALID': 'The Stop Loss on fill specifies an invalid price',
        'STOP_LOSS_ON_FILL_PRICE_PRECISION_EXCEEDED': 'The Stop Loss on fill specifies a price with more precision '
                                                      'than is allowed by the Order’s instrument',
        'STOP_LOSS_ON_FILL_TIME_IN_FORCE_MISSING': 'The Stop Loss on fill specified does not provide a TimeInForce',
        'STOP_LOSS_ON_FILL_TIME_IN_FORCE_INVALID': 'The Stop Loss on fill specifies an invalid TimeInForce',
        'STOP_LOSS_ON_FILL_GTD_TIMESTAMP_MISSING': 'The Stop Loss on fill specifies a GTD TimeInForce but does not'
                                                   ' provide a GTD timestamp',
        'STOP_LOSS_ON_FILL_GTD_TIMESTAMP_IN_PAST': 'The Stop Loss on fill specifies a GTD timestamp that is in the '
                                                   'past',
        'STOP_LOSS_ON_FILL_CLIENT_ORDER_ID_INVALID': 'The Stop Loss on fill client Order ID specified is invalid',
        'STOP_LOSS_ON_FILL_CLIENT_ORDER_TAG_INVALID': 'The Stop Loss on fill client Order tag specified is invalid',
        'STOP_LOSS_ON_FILL_CLIENT_ORDER_COMMENT_INVALID': 'The Stop Loss on fill client Order comment specified is '
                                                          'invalid',
        'STOP_LOSS_ON_FILL_TRIGGER_CONDITION_MISSING': 'The Stop Loss on fill specified does not provide a '
                                                       'TriggerCondition',
        'STOP_LOSS_ON_FILL_TRIGGER_CONDITION_INVALID': 'The Stop Loss on fill specifies an invalid TriggerCondition',
        'TRAILING_STOP_LOSS_ORDER_ALREADY_EXISTS': 'A Trailing Stop Loss Order for the specified Trade already exists',
        'TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_MISSING': 'The Trailing Stop Loss on fill specified does not '
                                                             'provide a distance',
        'TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_INVALID': 'The Trailing Stop Loss on fill distance is invalid',
        'TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_PRECISION_EXCEEDED': 'The Trailing Stop Loss on fill distance '
                                                                        'contains more precision than is allowed by '
                                                                        'the instrument',
        'TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_MAXIMUM_EXCEEDED': 'The Trailing Stop Loss on fill price distance '
                                                                      'exceeds the maximum allowed amount',
        'TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_MINIMUM_NOT_MET': 'The Trailing Stop Loss on fill price distance '
                                                                     'does not meet the minimum allowed amount',
        'TRAILING_STOP_LOSS_ON_FILL_TIME_IN_FORCE_MISSING': 'The Trailing Stop Loss on fill specified does not '
                                                            'provide a TimeInForce',
        'TRAILING_STOP_LOSS_ON_FILL_TIME_IN_FORCE_INVALID': 'The Trailing Stop Loss on fill specifies an invalid '
                                                            'TimeInForce',
        'TRAILING_STOP_LOSS_ON_FILL_GTD_TIMESTAMP_MISSING': 'The Trailing Stop Loss on fill TimeInForce is specified '
                                                            'as GTD but no GTD timestamp is provided',
        'TRAILING_STOP_LOSS_ON_FILL_GTD_TIMESTAMP_IN_PAST': 'The Trailing Stop Loss on fill GTD timestamp is in the '
                                                            'past',
        'TRAILING_STOP_LOSS_ON_FILL_CLIENT_ORDER_ID_INVALID': 'The Trailing Stop Loss on fill client Order ID '
                                                              'specified is invalid',
        'TRAILING_STOP_LOSS_ON_FILL_CLIENT_ORDER_TAG_INVALID': 'The Trailing Stop Loss on fill client Order tag '
                                                               'specified is invalid',
        'TRAILING_STOP_LOSS_ON_FILL_CLIENT_ORDER_COMMENT_INVALID': 'The Trailing Stop Loss on fill client Order '
                                                                   'comment specified is invalid',
        'TRAILING_STOP_LOSS_ORDERS_NOT_SUPPORTED': 'A client attempted to create either a Trailing Stop Loss order or '
                                                   'an order with a Trailing Stop Loss On Fill specified, which may '
                                                   'not yet be supported.',
        'TRAILING_STOP_LOSS_ON_FILL_TRIGGER_CONDITION_MISSING': 'The Trailing Stop Loss on fill specified does not '
                                                                'provide a TriggerCondition',
        'TRAILING_STOP_LOSS_ON_FILL_TRIGGER_CONDITION_INVALID': 'The Tailing Stop Loss on fill specifies an invalid '
                                                                'TriggerCondition',
        'CLOSE_TRADE_TYPE_MISSING': 'The request to close a Trade does not specify a full or partial close',
        'CLOSE_TRADE_PARTIAL_UNITS_MISSING': 'The request to close a Trade partially did not specify the number of '
                                             'units to close',
        'CLOSE_TRADE_UNITS_EXCEED_TRADE_SIZE': 'The request to partially close a Trade specifies a number of units '
                                               'that exceeds the current size of the given Trade',
        'CLOSEOUT_POSITION_DOESNT_EXIST': 'The Position requested to be closed out does not exist',
        'CLOSEOUT_POSITION_INCOMPLETE_SPECIFICATION': 'The request to closeout a Position was specified incompletely',
        'CLOSEOUT_POSITION_UNITS_EXCEED_POSITION_SIZE': 'A partial Position closeout request specifies a number of '
                                                        'units that exceeds the current Position',
        'CLOSEOUT_POSITION_REJECT': 'The request to closeout a Position could not be fully satisfied',
        'CLOSEOUT_POSITION_PARTIAL_UNITS_MISSING': 'The request to partially closeout a Position did not specify the '
                                                   'number of units to close.',
        'MARKUP_GROUP_ID_INVALID': 'The markup group ID provided is invalid',
        'POSITION_AGGREGATION_MODE_INVALID': 'The PositionAggregationMode provided is not supported/valid.',
        'ADMIN_CONFIGURE_DATA_MISSING': 'No configuration parameters provided',
        'MARGIN_RATE_INVALID': 'The margin rate provided is invalid',
        'MARGIN_RATE_WOULD_TRIGGER_CLOSEOUT': 'The margin rate provided would cause an immediate margin closeout',
        'ALIAS_INVALID': 'The account alias string provided is invalid',
        'CLIENT_CONFIGURE_DATA_MISSING': 'No configuration parameters provided',
        'MARGIN_RATE_WOULD_TRIGGER_MARGIN_CALL': 'The margin rate provided would cause the Account to enter a margin '
                                                 'call state.',
        'AMOUNT_INVALID': 'Funding is not possible because the requested transfer amount is invalid',
        'INSUFFICIENT_FUNDS': 'The Account does not have sufficient balance to complete the funding request',
        'AMOUNT_MISSING': 'Funding amount has not been specified',
        'FUNDING_REASON_MISSING': 'Funding reason has not been specified',
        'CLIENT_EXTENSIONS_DATA_MISSING': 'Neither Order nor Trade on Fill client extensions were provided for'
                                          ' modification',
        'REPLACING_ORDER_INVALID': 'The Order to be replaced has a different type than the replacing Order.',
        'REPLACING_TRADE_ID_INVALID': 'The replacing Order refers to a different Trade than the Order that is being '
                                      'replaced.'
    }


class TransactionType(Descriptor):
    """The possible types of a Transaction
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'CREATE': 'Account Create Transaction',
        'CLOSE': 'Account Close Transaction',
        'REOPEN': 'Account Reopen Transaction',
        'CLIENT_CONFIGURE': 'Client Configuration Transaction',
        'CLIENT_CONFIGURE_REJECT': 'Client Configuration Reject Transaction',
        'TRANSFER_FUNDS': 'Transfer Funds Transaction',
        'TRANSFER_FUNDS_REJECT': 'Transfer Funds Reject Transaction',
        'MARKET_ORDER': 'Market Order Transaction',
        'MARKET_ORDER_REJECT': 'Market Order Reject Transaction',
        'LIMIT_ORDER': 'Limit Order Transaction',
        'LIMIT_ORDER_REJECT': 'Limit Order Reject Transaction',
        'STOP_ORDER': 'Stop Order Transaction',
        'STOP_ORDER_REJECT': 'Stop Order Reject Transaction',
        'MARKET_IF_TOUCHED_ORDER': 'Market if Touched Order Transaction',
        'MARKET_IF_TOUCHED_ORDER_REJECT': 'Market if Touched Order Reject Transaction',
        'TAKE_PROFIT_ORDER': 'Take Profit Order Transaction',
        'TAKE_PROFIT_ORDER_REJECT': 'Take Profit Order Reject Transaction',
        'STOP_LOSS_ORDER': 'Stop Loss Order Transaction',
        'STOP_LOSS_ORDER_REJECT': 'Stop Loss Order Reject Transaction',
        'TRAILING_STOP_LOSS_ORDER': 'Trailing Stop Loss Order Transaction',
        'TRAILING_STOP_LOSS_ORDER_REJECT': 'Trailing Stop Loss Order Reject Transaction',
        'ORDER_FILL': 'Order Fill Transaction',
        'ORDER_CANCEL': 'Order Cancel Transaction',
        'ORDER_CANCEL_REJECT': 'Order Cancel Reject Transaction',
        'ORDER_CLIENT_EXTENSIONS_MODIFY': 'Order Client Extensions Modify Transaction',
        'ORDER_CLIENT_EXTENSIONS_MODIFY_REJECT': 'Order Client Extensions Modify Reject Transaction',
        'TRADE_CLIENT_EXTENSIONS_MODIFY': 'Trade Client Extensions Modify Transaction',
        'TRADE_CLIENT_EXTENSIONS_MODIFY_REJECT': 'Trade Client Extensions Modify Reject Transaction',
        'MARGIN_CALL_ENTER': 'Margin Call Enter Transaction',
        'MARGIN_CALL_EXTEND': 'Margin Call Extend Transaction',
        'MARGIN_CALL_EXIT': 'Margin Call Exit Transaction',
        'DELAYED_TRADE_CLOSURE': 'Delayed Trade Closure Transaction',
        'DAILY_FINANCING': 'Daily Financing Transaction',
        'RESET_RESETTABLE_PL': 'Reset Resettable PL Transaction'
    }



import inspect

import pytest
from async_v20.definitions import types
from async_v20.definitions.descriptors.base import Descriptor
from async_v20.definitions.helpers import assign_descriptors
# from async_v20.definitions.helpers import async_flatten_dict
from async_v20.definitions.helpers import create_signature
from async_v20.definitions.helpers import flatten_dict

nested_dict = {'a': {'b': 2, 'c': {'d': 4}}}
flattened_dict = {'a_b': 2, 'a_c_d': 4}


def test_flatten_dict():
    result = flatten_dict(nested_dict, delimiter='_')
    assert result == flattened_dict


@pytest.mark.parametrize('cls', map(lambda x: getattr(types, x), types.__all__))
def test_create_signature(cls):
    # Create the signature
    result = create_signature(cls)
    # Ensure result is a Signature
    assert type(result) == inspect.Signature
    # Create names to assert against
    parameter_names = [name for name in result.parameters.keys()]
    # Ensure all signature arguments are lowercase
    assert all(map(lambda x: x.islower(), parameter_names))

    # Create a dict of all the default arguments
    default_parameters = dict([(cls.instance_attributes[key], value.default)
                               for key, value in cls._schema.items()
                               if not value.default == inspect._empty])
    # Create a dict of all the required arguments. Use the default value if there is one
    required_parameters = dict([(cls.instance_attributes[key], 'TEST_PARAMETER')
                                for key, value in cls._schema.items()
                                if value.required])

    # Ensure error occurs when required parameters are missing
    if required_parameters:
        with pytest.raises(TypeError):
            bound = result.bind()

    kwargs = required_parameters
    kwargs.update(default_parameters)
    bound = result.bind(*(None,),**kwargs)
    bound.apply_defaults()
    # Ensure that the default parameters are assigned correctly
    assert all(map(lambda x: bound.arguments[x[0]] == x[1], default_parameters.items()))
    # Ensure that the required parameters
    assert all(map(lambda x: bound.arguments[x[0]] == x[1], required_parameters.items()))


@pytest.mark.parametrize('cls', map(lambda x: getattr(types, x), types.__all__))
def test_assign_descriptors(cls):
    descriptors = dict([(cls.instance_attributes[attr], schema_value.typ)
                   for attr, schema_value
                   in cls._schema.items()
                   if issubclass(schema_value.typ, Descriptor)])
    cls = assign_descriptors(cls)

    assert all(map(lambda x: hasattr(cls, x), descriptors.keys()))
    assert all(map(lambda x: type(getattr(cls, x[0])) == x[1], descriptors.items()))



from ..data.json_data import GETAccountID_response
from async_v20.definitions.types import Account

def test_account_builds_from_dict():
    account_instance = Account(**GETAccountID_response['account'])
    assert type(account_instance) == Account


from .base import Interface
from .decorators import endpoint
from ..definitions.types import TransactionID
from ..endpoints.annotations import FromTime
from ..endpoints.annotations import PageSize
from ..endpoints.annotations import ToTime
from ..endpoints.annotations import Type
from ..endpoints.annotations import FromTransactionID
from ..endpoints.annotations import ToTransactionID
from ..endpoints.transaction import *

__all__ = ['TransactionInterface']


class TransactionInterface(Interface):
    @endpoint(GETTransactions)
    def list_transactions(self, from_time: FromTime, to_time: ToTime, page_size: PageSize,
                          type_: Type):
        """
        Get a list of Transactions pages that satisfy a time-based Transaction
        query.

        Args:
            from_time:
                The starting time (inclusive) of the time range for the
                Transactions being queried.
            to_time:
                The ending time (inclusive) of the time range for the
                Transactions being queried.
            page_size:
                The number of Transactions to include in each page of the
                results.
            type_:
                A filter for restricting the types of Transactions to retrieve.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETTransactionID)
    def get_transactions(self, transaction_id: TransactionID):
        """
        Get the details of a single Account Transaction.

        Args:
            transaction_id:
                A Transaction ID

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETIDrange)
    def transaction_range(self, from_transaction: FromTransactionID, to_transaction: ToTransactionID, type_: Type):
        """
        Get a range of Transactions for an Account based on the Transaction
        IDs.

        Args:
            from_transaction:
                The starting Transaction ID (inclusive) to fetch.
            to_transaction:
                The ending Transaction ID (inclusive) to fetch.
            type_:
                The filter that restricts the types of Transactions to
                retrieve.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETSinceID)
    def since_transaction(self, transaction_id: TransactionID):
        """
        Get a range of Transactions for an Account starting at (but not
        including) a provided Transaction ID.

        Args:
            transaction_id:
                The ID of the last Transaction fetched. This query will return
                all Transactions newer than the TransactionID.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETTransactionsStream)
    def stream_transactions(self):
        """
        Get a stream of Transactions for an Account starting from when the
        request is made.

        Args:

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass



from .annotations import *
from .base import EndPoint, Path
from ..definitions.descriptors import *
from ..definitions.metaclass import Array
from ..definitions.types import *

__all__ = ['GETPositions', 'GETOpenPositions', 'GETPositionsInstrument', 'PUTPositionsInstrumentClose']


class GETPositions(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/positions')

    # description of endpoint
    description = 'List all Positions for an Account. The Positions returned are for every ' \
                  'instrument that has had a position during the lifetime of an the Account.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {200: {'positions': Array(Position), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class GETOpenPositions(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/openPositions')

    # description of endpoint
    description = 'List all open Positions for an Account. An open Position is a ' \
                  'Position in an Account that currently has a Trade opened for it.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {200: {'positions': Array(Position), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class GETPositionsInstrument(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/positions/', InstrumentName)

    # description of endpoint
    description = 'Get the details of a single Instrument’s Position in an Account. The Position may by open or not.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'instrument', 'located': 'path', 'type': InstrumentName, 'description': 'InstrumentName'},
    ]

    # valid responses
    responses = {200: {'position': Position, 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class PUTPositionsInstrumentClose(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'PUT'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/positions/', InstrumentName, '/close')

    # description of endpoint
    description = 'Closeout the open Position for a specific instrument in an Account.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'instrument', 'located': 'path', 'type': InstrumentName, 'description': 'InstrumentName'},
    ]

    # valid responses
    responses = {
        200: {'longOrderCreateTransaction': MarketOrderTransaction, 'longOrderFillTransaction': OrderFillTransaction,
              'longOrderCancelTransaction': OrderCancelTransaction,
              'shortOrderCreateTransaction': MarketOrderTransaction, 'shortOrderFillTransaction': OrderFillTransaction,
              'shortOrderCancelTransaction': OrderCancelTransaction, 'relatedTransactionIDs': Array(TransactionID),
              'lastTransactionID': TransactionID},
        400: {'longOrderRejectTransaction': MarketOrderRejectTransaction,
              'shortOrderRejectTransaction': MarketOrderRejectTransaction,
              'relatedTransactionIDs': Array(TransactionID), 'lastTransactionID': TransactionID, 'errorCode': str,
              'errorMessage': str},
        404: {'longOrderRejectTransaction': MarketOrderRejectTransaction,
              'shortOrderRejectTransaction': MarketOrderRejectTransaction,
              'relatedTransactionIDs': Array(TransactionID), 'lastTransactionID': TransactionID, 'errorCode': str,
              'errorMessage': str}}

    # error msgs'
    error = (401, 405)

    # TODO test longunits and short units need to default to 'ALL'
    # json schema representation
    request_schema = {'longUnits': LongUnits, 'longClientExtensions': ClientExtensions, 'shortUnits': ShortUnits,
                      'shortClientExtensions': ClientExtensions}



from .annotations import *
from .base import EndPoint, Path
from ..definitions.descriptors import *
from ..definitions.metaclass import Array
from ..definitions.types import *

__all__ = ['GETPricing', 'GETPricingStream']


class GETPricing(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/pricing')

    # description of endpoint
    description = 'Get pricing information for a specified list of Instruments within an Account.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'instruments', 'located': 'query', 'type': Instruments,
         'description': 'List of InstrumentName (csv)'},
        {'name': 'since', 'located': 'query', 'type': DateTime, 'description': 'DateTime'},
        {'name': 'includeUnitsAvailable', 'located': 'query', 'type': IncludeUnitsAvailable, 'description': 'boolean'}
    ]

    # valid responses
    responses = {200: {'prices': Array(Price), 'time': DateTime}}

    # error msgs'
    error = (400, 401, 404, 405)


class GETPricingStream(EndPoint):
    # host to use for this endpoint
    host = 'STREAM'

    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/pricing/stream')

    # description of endpoint
    description = '''Get a stream of Account Prices starting from when the request is made.
    This pricing stream does not include every single price created for the Account, 
    but instead will provide at most 4 prices per second (every 250 milliseconds) 
    for each instrument being requested. If more than one price is created for an instrument 
    during the 250 millisecond window, only the price in effect at the end of the window is sent.
    This means that during periods of rapid price movement, subscribers to this stream will 
    not be sent every price. Pricing windows for different connections to the price stream 
    are not all aligned in the same way (i.e.they are not all aligned to the top of the second).
    This means that during periods of rapid price movement, different subscribers may observe 
    different prices depending on their alignment.'''

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'instruments', 'located': 'query', 'type': Instruments,
         'description': 'List of InstrumentName (csv)'},
        {'name': 'snapshot', 'located': 'query', 'type': Snapshot, 'description': 'boolean'},
    ]

    # valid responses
    responses = {200: {'PRICE': Price, 'HEARTBEAT': PricingHeartbeat}}

    # error msgs'
    error = (400, 401, 404, 405)



from .annotations import *
from .base import EndPoint, Path
from ..definitions.types import *
from .annotations import UserSpecifier, Authorization

__all__ = ['GETUserSpecifier', 'GETExternalInfo']


class GETUserSpecifier(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/users/', UserSpecifier)

    # description of endpoint
    description = 'Fetch the user information for the specified user.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'string'},
        {'name': 'userSpecifier', 'located': 'query', 'type': UserSpecifier, 'description': 'The User Specifier'},
    ]

    # valid responses
    responses = {200: {'userInfo': UserInfo}}

    # error msgs'
    error = [401, 403, 405]


class GETExternalInfo(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/users/', UserSpecifier, '/externalInfo')

    # description of endpoint
    description = 'Fetch the externally-available user information for the specified user.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'string'},
        {'name': 'userSpecifier', 'located': 'query', 'type': UserSpecifier, 'description': 'The User Specifier'},
    ]

    # valid responses
    responses = {200: {'userInfo': UserInfoExternal}}

    # error msgs'
    error = (401, 403, 405)



from .base import Interface
from .decorators import endpoint
from ..endpoints.annotations import UserSpecifier
from ..endpoints.user import *

__all__ = ['UserInterface']


class UserInterface(Interface):
    @endpoint(GETUserSpecifier)
    def get_user_info(self, user_specifier: UserSpecifier):
        """
        Fetch the user information for the specified user. This endpoint is
        intended to be used by the user thyself to obtain their own
        information.

        Args:
            user_specifier:
                The User Specifier

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETExternalInfo)
    def get_external_user_info(self, user_specifier: UserSpecifier):
        """
        Fetch the externally-available user information for the specified user.
        This endpoint is intended to be used by 3rd parties that have been
        authorized by a user to view their personal information.

        Args:
            user_specifier:
                The User Specifier

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass



"""Module that defines the behaviour of the exposed client method calls by using decorators
"""
from functools import wraps
from inspect import signature

from .helpers import create_request_kwargs
from .helpers import make_args_optional
from .parser import parse_response


async def _serial_request_async_generator():
    self, request_args, endpoint = yield
    while True:
        request = await self.request.asend(None)
        response = request(**request_args)
        self, request_args, endpoint = yield await parse_response(self, response, endpoint)


def endpoint(endpoint, serial=False):
    """Define a method call to be exposed to the user"""

    if serial:
        serial_request = _serial_request_async_generator()
        endpoint.initialized = False

    def wrapper(method):
        """Take the wrapped method and return a coroutine"""

        method.endpoint = endpoint

        sig = make_args_optional(signature(method))

        @wraps(method)
        async def serial_wrap(self, *args, **kwargs):
            try:
                await self.initialize()
            except ValueError:
                pass

            if not endpoint.initialized:
                await serial_request.asend(None)
                endpoint.initialized = True

            request_args = create_request_kwargs(self, endpoint, sig, *args, **kwargs)

            return await serial_request.asend((self, request_args, endpoint))

        @wraps(method)
        async def parallel_wrap(self, *args, **kwargs):
            try:
                await self.initialize()
            except ValueError:
                pass

            request_args = create_request_kwargs(self, endpoint, sig, *args, **kwargs)

            request = await self.request.asend(None)

            response = request(**request_args)

            return await parse_response(self, response, endpoint)

        serial_wrap.__signature__ = sig

        parallel_wrap.__signature__ = sig

        return {True: serial_wrap, False: parallel_wrap}[serial]

    return wrapper


def add_signature(class_obj):
    sig = signature(class_obj.__init__)

    def wrapper(func):
        @wraps(func)
        def wrap(*args, **kwargs):
            return func(*args, **kwargs)

        wrap.__signature__ = sig
        return wrap

    return wrapper




class Response(dict):

    def json_dict(self):

        def value_to_json_dict(value):
            try:
                # If value is an Model object
                result = value.json_dict()
            except AttributeError:
                if isinstance(value, list):
                    # If value is an Array of Model objects
                    result = [obj.json_dict() for obj in value]
                else:
                    result = value
            return result

        return {key: value_to_json_dict(value) for key, value in self.items()}


from typing import List

from ..helpers import sleep


class Path(object):
    """capture the construction of a path"""

    def __init__(self, *path):
        self.path = path

    def __call__(self, arguments: dict, default: dict):
        def lookup(segment):
            try:
                result = arguments[segment]
            except KeyError:
                result = default[segment]
            return result

        return ''.join([segment if isinstance(segment, str)
                        else lookup(segment)
                        for segment in self.path])


class EndPoint(object):
    """Base object representation of an endpoint"""

    # Default host to use unless otherwise specified by a derived class
    host = 'REST'

    # Path needs to be a Path object
    path = Path

    # the HTTP verb to use for this endpoint
    method = str

    # description of endpoint
    description = str

    # parameters required to send to endpoint
    parameters = List[dict]

    # valid responses
    responses = List[dict]

    # error msgs'
    error = List[str]

    # json format the data body as per the response_schema below
    request_schema = dict()



from .base import Interface
from .decorators import endpoint
from ..definitions.types import InstrumentName
from ..endpoints.annotations import LongClientExtensions
from ..endpoints.annotations import LongUnits
from ..endpoints.annotations import ShortClientExtensions
from ..endpoints.annotations import ShortUnits
from ..endpoints.position import *

__all__ = ['PositionInterface']


class PositionInterface(Interface):
    @endpoint(GETPositions)
    def list_positions(self):
        """
        List all Positions for an Account. The Positions returned are for every
        instrument that has had a position during the lifetime of an the
        Account.

        Args:


        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETOpenPositions)
    def list_open_positions(self):
        """
        List all open Positions for an Account. An open Position is a Position
        in an Account that currently has a Trade opened for it.

        Args:


        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETPositionsInstrument)
    def get_positions(self, instrument: InstrumentName):
        """
        Get the details of a single Instrument's Position in an Account. The
        Position may by open or not.

        Args:

            instrument:
                Name of the Instrument

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(PUTPositionsInstrumentClose)
    def close_position(self, instrument: InstrumentName, long_units: LongUnits,
                       long_client_extensions: LongClientExtensions,
                       short_units: ShortUnits, short_client_extensions: ShortClientExtensions):
        """
        Closeout the open Position for a specific instrument in an Account.

        Args:

            instrument:
                Name of the Instrument
            long_units:
                Indication of how much of the long Position to closeout. Either
                the string "ALL", the string "NONE", or a DecimalNumber
                representing how many units of the long position to close using
                a PositionCloseout MarketOrder. The units specified must always
                be positive.
            long_client_extensions:
                The client extensions to add to the MarketOrder used to close
                the long position.
            short_units:
                Indication of how much of the short Position to closeout.
                Either the string "ALL", the string "NONE", or a DecimalNumber
                representing how many units of the short position to close
                using a PositionCloseout MarketOrder. The units specified must
                always be positive.
            short_client_extensions:
                The client extensions to add to the MarketOrder used to close
                the short position.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass



from ..definitions.types import ClientExtensions
from ..definitions.types import DateTime
from ..definitions.types import TransactionID


class Authorization(str):
    pass


class Instruments(str):
    pass


class Alias(str):
    pass


class Count(int):
    pass


class Smooth(object):
    pass  # bool


class includeFirst(object):
    pass  # bool


class DailyAlignment(int):
    pass


class AlignmentTimezone(str):
    pass


class Ids(str):
    pass


class LongUnits(str):
    pass


# this needs to default to 'ALL'

class ShortUnits(str):
    pass


# this also needs to default to 'ALL'

class IncludeUnitsAvailable(object):
    pass  # bool


class Snapshot(object):
    pass  # bool


class PageSize(int):
    pass


class Type(str):
    pass


class UserSpecifier(str):
    pass


class FromTime(DateTime):
    pass


class ToTime(DateTime):
    pass


class TradeClientExtensions(ClientExtensions):
    pass


class LongClientExtensions(ClientExtensions):
    pass


class ShortClientExtensions(ClientExtensions):
    pass


class Units(str):
    pass


class LastTransactionID(TransactionID):
    pass

class FromTransactionID(TransactionID):
    pass

class ToTransactionID(TransactionID):
    pass


from aiohttp import web

routes = [
    ('GET', '/v3/accounts'),
    ('GET', '/v3/accounts/TEST_ID'),
    ('GET', '/v3/accounts/TEST_ID/summary'),
    ('GET', '/v3/accounts/TEST_ID/instruments'),
    ('PATCH', '/v3/accounts/TEST_ID/configuration'),
    ('GET', '/v3/accounts/TEST_ID/changes'),
    ('GET', '/v3/instruments/0/candles'),
    ('POST', '/v3/accounts/TEST_ID/orders'),
    ('GET', '/v3/accounts/TEST_ID/orders'),
    ('GET', '/v3/accounts/TEST_ID/pendingOrders'),
    ('GET', '/v3/accounts/TEST_ID/orders/0'),
    ('PUT', '/v3/accounts/TEST_ID/orders/0'),
    ('PUT', '/v3/accounts/TEST_ID/orders/0/cancel'),
    ('PUT', '/v3/accounts/TEST_ID/orders/0/clientExtensions'),
    ('GET', '/v3/accounts/TEST_ID/positions'),
    ('GET', '/v3/accounts/TEST_ID/openPositions'),
    ('GET', '/v3/accounts/TEST_ID/positions/0'),
    ('PUT', '/v3/accounts/TEST_ID/positions/0/close'),
    ('GET', '/v3/accounts/TEST_ID/pricing'),
    ('GET', '/v3/accounts/TEST_ID/pricing/stream'),
    ('GET', '/v3/accounts/TEST_ID/trades'),
    ('GET', '/v3/accounts/TEST_ID/openTrades'),
    ('GET', '/v3/accounts/TEST_ID/trades/0'),
    ('PUT', '/v3/accounts/TEST_ID/trades/0/close'),
    ('PUT', '/v3/accounts/TEST_ID/trades/0/clientExtensions'),
    ('PUT', '/v3/accounts/TEST_ID/trades/0/orders'),
    ('GET', '/v3/accounts/TEST_ID/transactions'),
    ('GET', '/v3/accounts/TEST_ID/transactions/0'),
    ('GET', '/v3/accounts/TEST_ID/transactions/idrange'),
    ('GET', '/v3/accounts/TEST_ID/transactions/sinceid'),
    ('GET', '/v3/accounts/TEST_ID/transactions/stream'),
    ('GET', '/v3/users/0'),
    ('GET', '/v3/users/0/externalInfo')
]


def echo(request):
    print()
    return web.Response(body=request.body,
                        status=200,
                        reason='ECHO',
                        headers=request.headers,
                        content_type=request.content_type)


app = web.Application()

route_lookup = {'GET': app.router.add_get,
                'PUT': app.router.add_put,
                'PATCH': app.router.add_patch,
                'POST': app.router.add_post}

for route in routes:
    route_lookup[route[0]](route[1], echo)

web.run_app(app)


import pytest
from async_v20.client import OandaClient
import os
import re
from async_v20.endpoints.annotations import Authorization


@pytest.fixture()
def client():
    oanda_client = OandaClient()
    yield oanda_client
    del oanda_client

def test_oanda_client_has_correct_version(client):
    try:
        path = os.path.abspath(r'../../async_v20/setup.py')
        print(path)
        with open(path, 'r') as f:
            setup = f.read()
    except FileNotFoundError as e:
        print(e)
    else:
        setup_version = re.findall(r"(?<=version\s=\s').*?(?='\n)", setup)[0]
        assert setup_version == client.version

def test_oanda_client_finds_token():
    os.environ['OANDA_TOKEN'] = 'test_environ_token'
    client = OandaClient()
    assert client.default_parameters[Authorization] == 'Bearer test_environ_token'

def test_oanda_client_accepts_token():
    client = OandaClient(token='test_token')
    assert client.default_parameters[Authorization] == 'Bearer test_token'

def test_oanda_client_has_application_name(client):
    assert client.application == 'async_v20'

def test_oanda_client_constructs_url(client):
    assert client.hosts['REST'](path='test').human_repr() == 'https://api-fxpractice.oanda.com:443/test'
    assert client.hosts['STREAM'](path='test').human_repr() == 'https://stream-fxpractice.oanda.com/test'



import asyncio

from async_v20 import OandaClient


async def poll_account(poll_interval=6):
    async with OandaClient() as client:
        while True:
            result = await client.account_changes()
            print(result)
            await asyncio.sleep(poll_interval)


loop = asyncio.get_event_loop()
response = loop.run_until_complete(poll_account())



from .annotations import *
from .base import EndPoint, Path
from ..definitions.descriptors import *
from ..definitions.metaclass import Array
from ..definitions.types import *

__all__ = ['GETInstrumentsCandles', 'GETInstrumentOrderBook', 'GETInstrumentsPositionBook']


class GETInstrumentsCandles(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/instruments/', InstrumentName, '/candles')

    # description of endpoint
    description = 'Fetch candlestick data for an instrument.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'instrument', 'located': 'path', 'type': InstrumentName, 'description': 'InstrumentName'},
        {'name': 'price', 'located': 'query', 'type': Price, 'description': 'str'},
        {'name': 'granularity', 'located': 'query', 'type': CandlestickGranularity,
         'description': 'CandlestickGranularity'},
        {'name': 'count', 'located': 'query', 'type': Count, 'description': 'int'},
        {'name': 'from', 'located': 'query', 'type': FromTime, 'description': 'DateTime'},
        {'name': 'to', 'located': 'query', 'type': ToTime, 'description': 'DateTime'},
        {'name': 'smooth', 'located': 'query', 'type': Smooth, 'description': 'boolean'},
        {'name': 'includeFirst', 'located': 'query', 'type': includeFirst, 'description': 'boolean'},
        {'name': 'dailyAlignment', 'located': 'query', 'type': DailyAlignment, 'description': 'int'},
        {'name': 'alignmentTimezone', 'located': 'query', 'type': AlignmentTimezone, 'description': 'str'},
        {'name': 'weeklyAlignment', 'located': 'query', 'type': WeeklyAlignment,
         'description': 'WeeklyAlignment'},
    ]

    # valid responses
    responses = {
        200: {'instrument': InstrumentName, 'granularity': CandlestickGranularity, 'candles': Array(Candlestick)}}

    # error msgs'
    error = (400, 401, 404, 405)


class GETInstrumentOrderBook(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/instruments/', Instrument, '/orderBook')

    # description of endpoint
    description = 'Fetch a gzip compressed order book for an instrument.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'instrument', 'located': 'path', 'type': InstrumentName, 'description': 'InstrumentName'},
        {'name': 'time', 'located': 'query', 'type': DateTime, 'description': 'DateTime'},
    ]

    # valid responses
    responses = {200: {'orderBook': OrderBook}}

    # error msgs'
    error = (400, 401, 404, 405)


class GETInstrumentsPositionBook(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/instruments/', Instrument, '/positionBook')

    # description of endpoint
    description = 'Fetch a gzip compressed position book for an instrument.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'instrument', 'located': 'path', 'type': InstrumentName, 'description': 'InstrumentName'},
        {'name': 'time', 'located': 'query', 'type': DateTime, 'description': 'DateTime'},
    ]

    # valid responses
    responses = {200: {'positionBook': PositionBook}}

    # error msgs'
    error = (400, 401, 404, 405)



from .account import *
from .instrument import *
from .order import *
from .position import *
from .pricing import *
from .trade import *
from .transaction import *
from .user import *

__all__ = (account.__all__ +
           instrument.__all__ +
           order.__all__ +
           pricing.__all__ +
           trade.__all__ +
           transaction.__all__ +
           user.__all__)



import pytest
from async_v20 import interface


@pytest.mark.parametrize('interface_method', [method for cls in (getattr(interface, cls) for cls in interface.__all__)
                                              for method in cls.__dict__.values() if hasattr(method, 'endpoint')])
def test_interface_docstrings_have_all_arguments(interface_method):
    print(interface_method.__name__)
    print('SIGNATURE: ', interface_method.__signature__)
    sig_names = {name:value for name, value in interface_method.__signature__.parameters.items() if name != 'self'}
    # Call the wrapped method, we are testing the only thing inside method
    # in this test. So may as well do this to improve coverage
    interface_method.__wrapped__(*list(range(len(sig_names)+1)))
    for key in sig_names:
        assert key in interface_method.__doc__



from .base import Interface
from .decorators import endpoint, add_signature
from ..definitions.types import ClientExtensions
from ..definitions.types import InstrumentName
from ..definitions.types import LimitOrderRequest
from ..definitions.types import MarketIfTouchedOrderRequest
from ..definitions.types import MarketOrderRequest
from ..definitions.types import OrderID
from ..definitions.types import OrderRequest
from ..definitions.types import OrderSpecifier
from ..definitions.types import OrderStateFilter
from ..definitions.types import StopLossOrderReason
from ..definitions.types import StopLossOrderRequest
from ..definitions.types import StopOrderRequest
from ..definitions.types import TakeProfitOrderRequest
from ..definitions.types import TrailingStopLossOrderRequest
from ..endpoints.annotations import Count
from ..endpoints.annotations import Ids
from ..endpoints.annotations import TradeClientExtensions
from ..endpoints.order import *

__all__ = ['OrderInterface']


class OrderInterface(Interface):
    @endpoint(POSTOrders)
    def create_order(self, order: OrderRequest):
        """
        Create an Order for an Account

        Args:

            order:
                Specification of the Order to create

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETOrders)
    def list_orders(self, ids: Ids, state: OrderStateFilter, instrument: InstrumentName, count: Count,
                    before_id: OrderID):
        """
        Get a list of Orders for an Account

        Args:

            ids:
                List of Order IDs to retrieve
            state:
                The state to filter the requested Orders by
            instrument:
                The instrument to filter the requested orders by
            count:
                The maximum number of Orders to return
            before_id:
                The maximum Order ID to return. If not provided the most recent
                Orders in the Account are returned

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETPendingOrders)
    def list_pending_orders(self):
        """
        List all pending Orders in an Account

        Args:


        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETOrderSpecifier)
    def get_order(self, order_specifier: OrderSpecifier):
        """
        Get details for a single Order in an Account

        Args:

            order_specifier:
                The Order Specifier

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(PUTOrderSpecifier)
    def replace_order(self, order_specifier: OrderSpecifier, order: OrderRequest):
        """
        Replace an Order in an Account by simultaneously cancelling it and
        creating a replacement Order

        Args:

            order_specifier:
                The Order Specifier
            order:
                Specification of the replacing Order

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(PUTOrderSpecifierCancel)
    def cancel_order(self, order_specifier: OrderSpecifier):
        """
        Cancel a pending Order in an Account

        Args:

            order_specifier:
                The Order Specifier

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(PUTClientExtensions)
    def set_client_extensions(self, order_specifier: OrderSpecifier, client_extensions: ClientExtensions,
                              trade_client_extensions: TradeClientExtensions):
        """
        Update the Client Extensions for an Order in an Account. Do not set,
        modify, or delete clientExtensions if your account is associated with
        MT4.

        Args:

            order_specifier:
                The Order Specifier
            client_extensions:
                The Client Extensions to update for the Order. Do not set,
                modify, or delete clientExtensions if your account is
                associated with MT4.
            trade_client_extensions:
                The Client Extensions to update for the Trade created when the
                Order is filled. Do not set, modify, or delete clientExtensions
                if your account is associated with MT4.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @add_signature(MarketOrderRequest)
    def market_order(self, *args, **kwargs):
        """
        Shortcut to create a Market Order in an Account
        MarketOrderRequest
        Args:
            kwargs : The arguments to create a MarketOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.create_order(order=MarketOrderRequest(*args, **kwargs))

    @add_signature(LimitOrderRequest)
    def limit_order(self, *args, **kwargs):
        """
        Shortcut to create a Limit Order in an Account

        Args:
            kwargs : The arguments to create a LimitOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.create_order(order=LimitOrderRequest(*args, **kwargs))

    @add_signature(LimitOrderRequest)
    def limit_replace_order(self, order_id, *args, **kwargs):
        """
        Shortcut to replace a pending Limit Order in an Account

        Args:
            order_id : The ID of the Limit Order to replace
            kwargs : The arguments to create a LimitOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.replace_order(order_id, order=LimitOrderRequest(*args, **kwargs))

    @add_signature(StopOrderRequest)
    def stop_order(self, *args, **kwargs):
        """
        Shortcut to create a Stop Order in an Account

        Args:
            kwargs : The arguments to create a StopOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.create_order(order=StopOrderRequest(*args, **kwargs))

    @add_signature(StopOrderRequest)
    def stop_replace_order(self, order_id, *args, **kwargs):
        """
        Shortcut to replace a pending Stop Order in an Account

        Args:
            order_id : The ID of the Stop Order to replace
            kwargs : The arguments to create a StopOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.replace_order(order_id, order=StopOrderRequest(*args, **kwargs))

    @add_signature(MarketIfTouchedOrderRequest)
    def market_if_touched_order(self, *args, **kwargs):
        """
        Shortcut to create a MarketIfTouched Order in an Account

        Args:
            kwargs : The arguments to create a MarketIfTouchedOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.create_order(order=MarketIfTouchedOrderRequest(*args, **kwargs))

    @add_signature(MarketIfTouchedOrderRequest)
    def market_if_touched_replace_order(self, order_id, *args, **kwargs):
        """
        Shortcut to replace a pending MarketIfTouched Order in an Account

        Args:
            order_id : The ID of the MarketIfTouched Order to replace
            kwargs : The arguments to create a MarketIfTouchedOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.replace_order(order_id, order=MarketIfTouchedOrderRequest(*args, **kwargs))

    @add_signature(TakeProfitOrderRequest)
    def take_profit_order(self, *args, **kwargs):
        """
        Shortcut to create a Take Profit Order in an Account

        Args:
            kwargs : The arguments to create a TakeProfitOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.create_order(order=TakeProfitOrderRequest(*args, **kwargs))

    @add_signature(TakeProfitOrderRequest)
    def take_profit_replace_order(self, order_id, *args, **kwargs):
        """
        Shortcut to replace a pending Take Profit Order in an Account

        Args:
            order_id : The ID of the Take Profit Order to replace
            kwargs : The arguments to create a TakeProfitOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.replace_order(order_id, order=TakeProfitOrderRequest(*args, **kwargs))

    @add_signature(StopOrderRequest)
    def stop_loss_order(self, *args, **kwargs):
        """
        Shortcut to create a Stop Loss Order in an Account

        Args:
            kwargs : The arguments to create a StopLossOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.create_order(order=StopLossOrderRequest(*args, **kwargs))

    @add_signature(StopLossOrderReason)
    def stop_loss_replace_order(self, order_id, *args, **kwargs):
        """
        Shortcut to replace a pending Stop Loss Order in an Account

        Args:
            order_id : The ID of the Stop Loss Order to replace
            kwargs : The arguments to create a StopLossOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.replace_order(order_id, order=StopLossOrderRequest(*args, **kwargs))

    @add_signature(TrailingStopLossOrderRequest)
    def trailing_stop_loss_order(self, *args, **kwargs):
        """
        Shortcut to create a Trailing Stop Loss Order in an Account

        Args:
            kwargs : The arguments to create a TrailingStopLossOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.create_order(order=TrailingStopLossOrderRequest(*args, **kwargs))

    @add_signature(TrailingStopLossOrderRequest)
    def trailing_stop_loss_replace_order(self, order_id, *args, **kwargs):
        """
        Shortcut to replace a pending Trailing Stop Loss Order in an Account

        Args:
            order_id : The ID of the Take Profit Order to replace
            kwargs : The arguments to create a TrailingStopLossOrderRequest

        Returns:
            async_v20.interface.parser.Response containing the results from submitting
            the request
        """
        return self.replace_order(order_id, order=TrailingStopLossOrderRequest(*args, **kwargs))



from .annotations import *
from .base import EndPoint, Path
from ..definitions.descriptors import *
from ..definitions.metaclass import Array
from ..definitions.types import *

__all__ = ['POSTOrders', 'GETOrders', 'GETPendingOrders', 'GETOrderSpecifier', 'PUTOrderSpecifier',
           'PUTOrderSpecifierCancel', 'PUTClientExtensions']


class POSTOrders(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'POST'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/orders')

    # description of endpoint
    description = 'Create an Order for an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {201: {'orderCreateTransaction': Transaction, 'orderFillTransaction': OrderFillTransaction,
                       'orderCancelTransaction': OrderCancelTransaction, 'orderReissueTransaction': Transaction,
                       'orderReissueRejectTransaction': Transaction, 'relatedTransactionIDs': Array(TransactionID),
                       'lastTransactionID': TransactionID},
                 400: {'orderRejectTransaction': Transaction, 'relatedTransactionIDs': Array(TransactionID),
                       'lastTransactionID': TransactionID, 'errorCode': str, 'errorMessage': str},
                 404: {'orderRejectTransaction': Transaction, 'relatedTransactionIDs': Array(TransactionID),
                       'lastTransactionID': TransactionID, 'errorCode': str, 'errorMessage': str}}

    # error msgs'
    error = (401, 403, 405)

    # json schema representation
    request_schema = {'order': OrderRequest}


class GETOrders(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/orders')

    # description of endpoint
    description = 'Get a list of Orders for an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'ids', 'located': 'query', 'type': Ids,
         'description': 'List of OrderID (csv)'},
        {'name': 'state', 'located': 'query', 'type': OrderStateFilter, 'description': 'OrderStateFilter'},
        {'name': 'instrument', 'located': 'query', 'type': InstrumentName, 'description': 'InstrumentName'},
        {'name': 'count', 'located': 'query', 'type': Count, 'description': 'int'},
        {'name': 'beforeID', 'located': 'query', 'type': OrderID, 'description': 'OrderID'},
    ]

    # valid responses
    responses = {200: {'orders': Array(Order), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (400, 404, 405)


class GETPendingOrders(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/pendingOrders')

    # description of endpoint
    description = 'List all pending Orders in an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {200: {'orders': Array(Order), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class GETOrderSpecifier(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/orders/', OrderSpecifier)

    # description of endpoint
    description = 'Get details for a single Order in an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'orderSpecifier', 'located': 'path', 'type': OrderSpecifier, 'description': 'OrderSpecifier'},
    ]

    # valid responses
    responses = {200: {'order': Order, 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class PUTOrderSpecifier(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'PUT'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/orders/', OrderSpecifier)

    # description of endpoint
    description = 'Replace an Order in an Account by simultaneously cancelling it and creating a replacement Order'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'orderSpecifier', 'located': 'path', 'type': OrderSpecifier, 'description': 'OrderSpecifier'},
    ]

    # valid responses
    responses = {201: {'orderCancelTransaction': OrderCancelTransaction, 'orderCreateTransaction': Transaction,
                       'orderFillTransaction': OrderFillTransaction, 'orderReissueTransaction': Transaction,
                       'orderReissueRejectTransaction': Transaction,
                       'replacingOrderCancelTransaction': OrderCancelTransaction,
                       'relatedTransactionIDs': Array(TransactionID), 'lastTransactionID': TransactionID},
                 400: {'orderRejectTransaction': Transaction, 'relatedTransactionIDs': Array(TransactionID),
                       'lastTransactionID': TransactionID, 'errorCode': str, 'errorMessage': str},
                 404: {'orderCancelRejectTransaction': Transaction, 'relatedTransactionIDs': Array(TransactionID),
                       'lastTransactionID': TransactionID, 'errorCode': str, 'errorMessage': str}}

    # error msgs'
    error = (401, 405)

    # json schema representation
    request_schema = {'order': OrderRequest}


class PUTOrderSpecifierCancel(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'PUT'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/orders/', OrderSpecifier, '/cancel')

    # description of endpoint
    description = 'Cancel a pending Order in an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'orderSpecifier', 'located': 'path', 'type': OrderSpecifier, 'description': 'OrderSpecifier'},
    ]

    # valid responses
    responses = {200: {'orderCancelTransaction': OrderCancelTransaction, 'relatedTransactionIDs': Array(TransactionID),
                       'lastTransactionID': TransactionID},
                 400: {'orderCancelRejectTransaction': OrderCancelRejectTransaction,
                       'relatedTransactionIDs': Array(TransactionID), 'lastTransactionID': TransactionID,
                       'errorCode': str, 'errorMessage': str}}

    # error msgs'
    error = (401, 405)


class PUTClientExtensions(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'PUT'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/orders/', OrderSpecifier, '/clientExtensions')

    # description of endpoint
    description = 'Update the Client Extensions for an Order in an Account. Do not set, ' \
                  'modify, or delete clientExtensions if your account is associated with MT4.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'orderSpecifier', 'located': 'path', 'type': OrderSpecifier, 'description': 'OrderSpecifier'},
    ]

    # valid responses
    responses = {200: {'orderClientExtensionsModifyTransaction': OrderClientExtensionsModifyTransaction,
                       'lastTransactionID': TransactionID, 'relatedTransactionIDs': Array(TransactionID)},
                 400: {'orderClientExtensionsModifyRejectTransaction': OrderClientExtensionsModifyRejectTransaction,
                       'lastTransactionID': TransactionID, 'relatedTransactionIDs': Array(TransactionID),
                       'errorCode': str, 'errorMessage': str},
                 404: {'orderClientExtensionsModifyRejectTransaction': OrderClientExtensionsModifyRejectTransaction,
                       'lastTransactionID': TransactionID, 'relatedTransactionIDs': Array(TransactionID),
                       'errorCode': str, 'errorMessage': str}}

    # error msgs'
    error = (401, 405)

    # json schema representation
    request_schema = {'clientExtensions': ClientExtensions, 'tradeClientExtensions': ClientExtensions}



from functools import wraps

from .helpers import assign_descriptors
from .helpers import create_attribute
from .helpers import create_instance_attributes
from .helpers import create_json_attributes
from .helpers import create_signature
from .helpers import flatten_dict


class JSONArray(object):
    typ = None

    def __new__(cls, data):
        try:
            return [create_attribute(cls.typ, obj) for obj in data]
        except TypeError as e:
            msg = f'FAILED TO CREATE OBJECT: {cls.typ} FROM DATA: {data} DATA TYPE: {type(data)}'
            print(e.args)
            raise Exception(msg)


class Array(type):
    """Used to denote objects that are sent from OANDA in an array.
    Also used to correctly serialize objects.
    """
    def __new__(mcs, typ):
        return super().__new__(mcs, f'Array_{typ.__name__}', (JSONArray,), {'typ': typ})


class Dispatch(dict):
    """Keep track of the parent class a subclass is derived from"""

    def __init__(self, derived, **kwargs):
        # derived is the parent class
        self.derived = derived
        super().__init__(self, **kwargs)
        self.name = self.__class__.__name__

    def update(self, *args):
        for cls in args[0].values():
            cls._derived = self.derived
        super().update(*args)


def auto_assign(func, signature):
    @wraps(func)
    def __init__(self, *args, **kwargs):
        self._fields = []  # Would normally place this is the class. Didn't segment instance attrs though

        # This dict allow for camelCase and snake_case to be passed without error
        kwargs = {self.__class__.instance_attributes[key]: value for key, value in kwargs.items()}
        # Bind passed arguments to the generated signature for this class
        bound = signature.bind(self, *args, **kwargs)
        bound.apply_defaults()

        # When assigning attrs to the instance. The passed value is then passed to the argument's annotation
        # Which is basically type checking. In order to do this, we need a list of tuples with:
        # name :- name of attribute
        # annotation :- signature annotation
        # value :- passed value
        annotations = {attr: param.annotation for attr, param in signature.parameters.items()}
        arguments = [(attr, annotations[attr], value) for attr, value in bound.arguments.items()
                     if value is not None and attr != 'self']

        # Instantiate annotations with value and assign to instance
        for name, annotation, value in arguments:
            self._fields.append(name)
            attribute_value = create_attribute(annotation, value)
            setattr(self, name, attribute_value)

    # Update the signature of the class __init__
    __init__.__signature__ = signature
    return __init__


class ORM(type):
    instance_attributes = {}
    json_attributes = {}

    def __new__(mcs, *args, **kwargs):
        class_obj = super().__new__(mcs, *args, **kwargs)


        # async_v20 allows camelCase and snake_case to be used when instantiating objects.
        # arguments passed in camelCase are converted to snake_case prior to attribute assignment.
        # Signature arguments are also snake_case.
        # when serializing objects. Attributes must be converted to camelCase,
        # as per OANDA's specification.
        # To achieve this these two dictionary's are created:
        #   instance_attributes :- camelCase & snake_case keys -> snake_case values
        #   json_attributes :- snake_case keys -> camelCase values
        mcs.instance_attributes.update(create_instance_attributes(class_obj))
        mcs.json_attributes.update(create_json_attributes(class_obj))

        # Instrument the class' with descriptors corresponding to OANDA's definitions
        class_obj = assign_descriptors(class_obj)
        # assign the standard __init__ to correctly initialize objects from json
        class_obj.__init__ = auto_assign(class_obj.__init__, create_signature(class_obj))
        # This dictionary is used to create pandas.Series objects. The template ensures that all
        # like objects have same length Series allowing for them to be passed into a dataframe
        class_obj.template = dict.fromkeys(flatten_dict(class_obj._schema))

        # This attribute is used to keep track of subclasses for specialized creation
        class_obj._dispatch = Dispatch(class_obj)

        return class_obj



string = str
error_response = {'errorCode': string, 'errorMessage': string}
other_responses = {
    # A “400 Bad Request” reponse may be returned from the v20 REST API when the client has
    # provided invalid data to be processed.
    400: error_response,
    # A “401 Unauthorized” reponse may be returned from the v20 REST API when the endpoint being
    # accessed requires the client to be authenticated however the authentication token is invalid
    # or has not been provided.
    401: error_response,
    # A “403 Forbidden” response may be returned from the v20 REST API when the client has provided
    # a token that does not authorize them to perform the action implemented by the API endpoint.
    403: error_response,
    #A “404 Not Found” response may be returned from the v20 REST API when the client is attempting
    # to refer to an entity (Account, Trade, Order, Position, etc.) that does not exist.
    404: error_response,
    #A “405 Method Not Allowed” response may be returned from the v20 REST API when the client is
    # attempting to access an API endpoint with an HTTP method that is not supported.
    405: error_response,
    #A “416 Range Not Satisfiable” response may be returned from the v20 REST API when the client
    # has specified a range (time range or Transaction range) that is invalid or cannot be processed.
    416: error_response
}



from .annotations import *
from .base import EndPoint, Path
from ..definitions.descriptors import *
from ..definitions.metaclass import Array
from ..definitions.types import *

__all__ = ['GETTrades', 'GETOpenTrades', 'GETTradeSpecifier', 'PUTTradeSpecifierClose',
           'PUTTradeSpecifierClientExtensions', 'PUTTradesSpecifierOrders']


class GETTrades(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/trades')

    # description of endpoint
    description = 'Get a list of Trades for an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'ids', 'located': 'query', 'type': Ids,
         'description': 'List of TradeID (csv)'},
        {'name': 'state', 'located': 'query', 'type': TradeStateFilter, 'description': 'TradeStateFilter'},
        {'name': 'instrument', 'located': 'query', 'type': InstrumentName, 'description': 'InstrumentName'},
        {'name': 'count', 'located': 'query', 'type': Count, 'description': 'int'},
        {'name': 'beforeID', 'located': 'query', 'type': TradeID, 'description': 'TradeID'},
    ]

    # valid responses
    responses = {200: {'trades': Array(Trade), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class GETOpenTrades(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/openTrades')

    # description of endpoint
    description = 'Get the list of open Trades for an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {200: {'trades': Array(Trade), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class GETTradeSpecifier(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/trades/', TradeSpecifier)

    # description of endpoint
    description = 'Get the details of a specific Trade in an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'tradeSpecifier', 'located': 'path', 'type': TradeSpecifier, 'description': 'TradeSpecifier'},
    ]

    # valid responses
    responses = {200: {'trade': Trade, 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405)


class PUTTradeSpecifierClose(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'PUT'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/trades/', TradeSpecifier, '/close')

    # description of endpoint
    description = 'Close (partially or fully) a specific open Trade in an Account'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'tradeSpecifier', 'located': 'path', 'type': TradeSpecifier, 'description': 'TradeSpecifier'},
    ]

    # valid responses
    responses = {200: {'orderCreateTransaction': MarketOrderTransaction, 'orderFillTransaction': OrderFillTransaction,
                       'orderCancelTransaction': OrderCancelTransaction, 'relatedTransactionIDs': Array(TransactionID),
                       'lastTransactionID': TransactionID},
                 400: {'orderRejectTransaction': MarketOrderRejectTransaction, 'errorCode': str,
                       'errorMessage': str},
                 404: {'orderRejectTransaction': MarketOrderRejectTransaction, 'lastTransactionID': TransactionID,
                       'relatedTransactionIDs': Array(TransactionID), 'errorCode': str, 'errorMessage': str}}

    # error msgs'
    error = (401, 405)

    # TODO test to default to 'ALL'
    # json schema representation
    request_schema = {'units': Units}


class PUTTradeSpecifierClientExtensions(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'PUT'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/trades/', TradeSpecifier, '/clientExtensions')

    # description of endpoint
    description = 'Update the Client Extensions for a Trade. Do not add, update, or ' \
                  'delete the Client Extensions if your account is associated with MT4.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'tradeSpecifier', 'located': 'path', 'type': TradeSpecifier, 'description': 'TradeSpecifier'},
    ]

    # valid responses
    responses = {200: {'tradeClientExtensionsModifyTransaction': TradeClientExtensionsModifyTransaction,
                       'relatedTransactionIDs': Array(TransactionID), 'lastTransactionID': TransactionID},
                 400: {'tradeClientExtensionsModifyRejectTransaction': TradeClientExtensionsModifyRejectTransaction,
                       'lastTransactionID': TransactionID, 'relatedTransactionIDs': Array(TransactionID),
                       'errorCode': str, 'errorMessage': str},
                 404: {'tradeClientExtensionsModifyRejectTransaction': TradeClientExtensionsModifyRejectTransaction,
                       'lastTransactionID': TransactionID, 'relatedTransactionIDs': Array(TransactionID),
                       'errorCode': str, 'errorMessage': str}}

    # error msgs'
    error = (401, 405)

    # json schema representation
    request_schema = {'clientExtensions': ClientExtensions}


class PUTTradesSpecifierOrders(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'PUT'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/trades/', TradeSpecifier, '/orders')

    # description of endpoint
    description = 'Create, replace and cancel a Trade’s dependent Orders ' \
                  '(Take Profit, Stop Loss and Trailing Stop Loss) through the Trade itself'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'tradeSpecifier', 'located': 'path', 'type': TradeSpecifier, 'description': 'TradeSpecifier'},
    ]

    # valid responses
    responses = {200: {'takeProfitOrderCancelTransaction': OrderCancelTransaction,
                       'takeProfitOrderTransaction': TakeProfitOrderTransaction,
                       'takeProfitOrderFillTransaction': OrderFillTransaction,
                       'takeProfitOrderCreatedCancelTransaction': OrderCancelTransaction,
                       'stopLossOrderCancelTransaction': OrderCancelTransaction,
                       'stopLossOrderTransaction': StopLossOrderTransaction,
                       'stopLossOrderFillTransaction': OrderFillTransaction,
                       'stopLossOrderCreatedCancelTransaction': OrderCancelTransaction,
                       'trailingStopLossOrderCancelTransaction': OrderCancelTransaction,
                       'trailingStopLossOrderTransaction': TrailingStopLossOrderTransaction,
                       'relatedTransactionIDs': Array(TransactionID), 'lastTransactionID': TransactionID},
                 400: {'takeProfitOrderCancelRejectTransaction': OrderCancelRejectTransaction,
                       'takeProfitOrderRejectTransaction': TakeProfitOrderRejectTransaction,
                       'stopLossOrderCancelRejectTransaction': OrderCancelRejectTransaction,
                       'stopLossOrderRejectTransaction': StopLossOrderRejectTransaction,
                       'trailingStopLossOrderCancelRejectTransaction': OrderCancelRejectTransaction,
                       'trailingStopLossOrderRejectTransaction': TrailingStopLossOrderRejectTransaction,
                       'lastTransactionID': TransactionID, 'relatedTransactionIDs': Array(TransactionID),
                       'errorCode': str, 'errorMessage': str}}

    # error msgs'
    error = (401, 404, 405)

    # json schema representation
    request_schema = {'takeProfit': TakeProfitDetails, 'stopLoss': StopLossDetails,
                      'trailingStopLoss': TrailingStopLossDetails}



from inspect import isclass

import pytest
from async_v20.definitions import descriptors
from async_v20.definitions.descriptors.base import IncorrectValue, LengthError
from hypothesis.strategies import text, integers, floats, sampled_from

# Get All user defined class' in descriptor package
@pytest.mark.parametrize('descriptor', map(lambda x: getattr(descriptors,x), descriptors.__all__))
def test_all_correct_descriptor_values_can_be_assigned(descriptor):
    if isclass(descriptor):
        # Create a dummy class to assign attributes to
        class TestClass(object):
            attr = descriptor(name='_attr')

        test_class = TestClass()
        # Test that every key in values can be assigned
        if descriptor.values:
            for key in descriptor.values.keys():
                test_class.attr = key

                assert test_class.attr == key


# Get All user defined class' in descriptor package
@pytest.mark.parametrize('descriptor', map(lambda x: getattr(descriptors,x), descriptors.__all__))
def test_all_incorrect_descriptor_values_cannot_be_assigned(descriptor):
    if isclass(descriptor):
        # Create a dummy class to assign attributes to
        class TestClass(object):
            attr = descriptor(name='_attr')

        test_class = TestClass()
        # Test that every key in values can be assigned
        if descriptor.values:
            for key in descriptor.values.keys():
                with pytest.raises(IncorrectValue):
                    test_class.attr = (key + '_test')


# Get All user defined class' in descriptor package
@pytest.mark.parametrize('descriptor', map(lambda x: getattr(descriptors,x), descriptors.__all__))
def test_descriptor_values_have_length_checking(descriptor):
    if isclass(descriptor):
        # Create a dummy class to assign attributes to
        class TestClass(object):
            attr = descriptor(name='_attr')

        test_class = TestClass()
        # Test that every key in values can be assigned
        if descriptor.values:
            for key in descriptor.values.keys():
                with pytest.raises(IncorrectValue):
                    test_class.attr = (key + '_test')


# Get All user defined class' in descriptor package
@pytest.mark.parametrize('descriptor', map(lambda x: getattr(descriptors,x), descriptors.__all__))
def test_descriptors_enforce_length_checking(descriptor):
    if isclass(descriptor):
        # Create a dummy class to assign attributes to
        class TestClass(object):
            attr = descriptor(name='_attr')

        test_class = TestClass()
        # Test that every key in values can be assigned
        if descriptor.example:
            with pytest.raises(LengthError):
                test_class.attr = (descriptor.example + '_')


# Get All user defined class' in descriptor package
@pytest.mark.parametrize('descriptor', map(lambda x: getattr(descriptors,x), descriptors.__all__))
def test_descriptors_convert_type_to_float(descriptor):
    if isclass(descriptor) and descriptor.typ == float:
        # Create a dummy class to assign attributes to
        class TestClass(object):
            attr = descriptor(name='_attr')

        test_class = TestClass()
        # Test that when a str is passed to a descriptor that requires a float the descriptor will convert it to a float
        test_class.attr = '0.1234'
        assert type(test_class.attr) == float


data_gen = {str: sampled_from([text(), integers(), floats()]), int: sampled_from(['1','-1',3.5]), float: integers()}

@pytest.mark.parametrize('descriptor', map(lambda x: getattr(descriptors,x), descriptors.__all__))
def test_descriptors_return_correct_type_when_initialized_with_value(descriptor):
    typ = descriptor.typ
    gen = data_gen[typ]
    assert all(map(lambda x: type(x) == typ, (descriptor(gen.example()) for _ in range(50))))

def test_PriceValue_rounds_floats_to_the_correct_accuracy():
    class TestClass(object):
        attribute = descriptors.PriceValue(name='_' + 'attribute')

    test_class = TestClass()
    test_class.attribute = 0.123456
    assert test_class.attribute == 0.12346

def test_Unit_rounds_floats_to_the_correct_accuracy():
    class TestClass(object):
        attribute = descriptors.Unit(name='_' + 'attribute')

    test_class = TestClass()
    test_class.attribute = 0.6
    assert test_class.attribute == 1

def test_delattr():
    class TestClass(object):
        attribute = descriptors.Unit(name='_' + 'attribute')

    test_class = TestClass()
    test_class.attribute = 0.6
    assert hasattr(test_class, '_attribute')
    del test_class._attribute
    assert not hasattr(test_class, '_attribute')


from .base import Descriptor

__all__ = ['CancellableOrderType', 'OrderID', 'OrderPositionFill', 'OrderSpecifier', 'OrderState', 'OrderStateFilter',
           'OrderTriggerCondition', 'OrderType', 'TimeInForce']


class CancellableOrderType(Descriptor):
    """The type of the Order.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'LIMIT': 'A Limit Order',
        'STOP': 'A Stop Order',
        'MARKET_IF_TOUCHED': 'A Market-if-touched Order',
        'TAKE_PROFIT': 'A Take Profit Order',
        'STOP_LOSS': 'A Stop Loss Order',
        'TRAILING_STOP_LOSS': 'A Trailing Stop Loss Order'
    }


class OrderID(Descriptor):
    """The Order’s identifier, unique within the Order’s Account.
    """

    # Type checking
    typ = int

    # Correct syntax of value
    format_syntax = 'The string representation of the OANDA-assigned OrderID. ' \
                    'OANDA-assigned OrderIDs are positive integers, and are derived from the ' \
                    'TransactionID of the Transaction that created the Order.'
    # # Example of correct format
    # example = '1523'


class OrderPositionFill(Descriptor):
    """Specification of how Positions in the Account
    are modified when the Order is filled.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'OPEN_ONLY': 'When the Order is filled, only allow Positions to be opened or extended.',
        'REDUCE_FIRST': 'When the Order is filled, always fully reduce an existing Position '
                        'before opening a new Position.',
        'REDUCE_ONLY': 'When the Order is filled, only reduce an existing Position.',
        'DEFAULT': 'When the Order is filled, use REDUCE_FIRST behaviour for non-client hedging Accounts, '
                   'and OPEN_ONLY behaviour for client hedging Accounts.'
    }


class OrderSpecifier(Descriptor):
    """The specification of an Order as referred to by clients
    """

    # TODO Check with OANDA. spec = int. I believe it should be a str
    # Type checking
    typ = str

    # Correct syntax of value
    format_syntax = 'Either the Order’s OANDA-assigned OrderID or the Order’s client-provided ' \
                    'ClientID prefixed by the “@” symbol'
    # Example of correct format
    # example = '1523'


class OrderState(Descriptor):
    """The current state of the Order.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'PENDING': 'The Order is currently pending execution',
        'FILLED': 'The Order has been filled',
        'TRIGGERED': 'The Order has been triggered',
        'CANCELLED': 'The Order has been cancelled'
    }


class OrderStateFilter(Descriptor):
    """The state to filter the requested Orders by.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'PENDING': 'The Orders that are currently pending execution',
        'FILLED': 'The Orders that have been filled',
        'TRIGGERED': 'The Orders that have been triggered',
        'CANCELLED': 'The Orders that have been cancelled',
        'ALL': 'The Orders that are in any of the possible states listed above'
    }


class OrderTriggerCondition(Descriptor):
    """Specification of which price component should be used when determining if an
    Order should be triggered and filled. This allows Orders to be triggered based
    on the bid, ask, mid, default (ask for buy, bid for sell) or inverse (
    ask for sell, bid for buy) price depending on the desired behaviour. Orders are
    always filled using their default price component. This feature is only provided
    through the REST API. Clients who choose to specify a non-default trigger condition
    will not see it reflected in any of OANDA’s proprietary or partner trading platforms,
    their transaction history or their account statements. OANDA platforms always assume
    that an Order’s trigger condition is set to the default value when indicating the distance
    from an Order’s trigger price, and will always provide the default trigger condition
    when creating or modifying an Order.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'DEFAULT': 'Trigger an Order the “natural” way: '
                   'compare its price to the ask for long Orders and bid for short Orders.',
        'INVERSE': 'Trigger an Order the opposite of the “natural” way: '
                   'compare its price the bid for long Orders and ask for short Orders.',
        'BID': 'Trigger an Order by comparing its price to the bid regardless of whether it is long or short.',
        'ASK': 'Trigger an Order by comparing its price to the ask regardless of whether it is long or short.',
        'MID': 'Trigger an Order by comparing its price to the midpoint regardless of whether it is long or short.'
    }


class OrderType(Descriptor):
    """The type of the Order.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'MARKET': 'A Market Order',
        'LIMIT': 'A Limit Order',
        'STOP': 'A Stop Order',
        'MARKET_IF_TOUCHED': 'A Market-if-touched Order',
        'TAKE_PROFIT': 'A Take Profit Order',
        'STOP_LOSS': 'A Stop Loss Order',
        'TRAILING_STOP_LOSS': 'A Trailing Stop Loss Order'
    }


class TimeInForce(Descriptor):
    """The time-in-force of an Order. TimeInForce describes how long an Order
    should remain pending before being automatically cancelled by the execution system.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'GTC': 'The Order is “Good unTil Cancelled”',
        'GTD': 'The Order is “Good unTil Date” and will be cancelled at the provided time',
        'GFD': 'The Order is “Good For Day” and will be cancelled at 5pm New York time',
        'FOK': 'The Order must be immediately “Filled Or Killed”',
        'IOC': 'The Order must be “Immediatedly paritally filled Or Cancelled”'
    }



from .base import Interface
from .decorators import endpoint
from ..endpoints.annotations import *
from ..endpoints.pricing import *

__all__ = ['PricingInterface']


class PricingInterface(Interface):
    @endpoint(GETPricing)
    def get_pricing(self, instruments: Instruments, since: DateTime, include_units_available: IncludeUnitsAvailable):
        """
        Get pricing information for a specified list of Instruments within an
        Account.

        Args:

            instruments:
                List of Instruments to get pricing for.
            since:
                Date/Time filter to apply to the returned prices. Only prices
                with a time later than this filter will be provided.
            include_units_available:
                Flag that enables the inclusion of the unitsAvailable field in
                the returned Price objects.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass

    @endpoint(GETPricingStream)
    def stream_pricing(self, instruments: Instruments, snapshot: Snapshot):
        """
        Get a stream of Account Prices starting from when the request is made.
        This pricing stream does not include every single price created for the
        Account, but instead will provide at most 4 prices per second (every
        250 milliseconds) for each instrument being requested. If more than one
        price is created for an instrument during the 250 millisecond window,
        only the price in effect at the end of the window is sent. This means
        that during periods of rapid price movement, subscribers to this stream
        will not be sent every price. Pricing windows for different connections
        to the price stream are not all aligned in the same way (i.e. they are
        not all aligned to the top of the second). This means that during
        periods of rapid price movement, different subscribers may observe
        different prices depending on their alignment.

        Args:

            instruments:
                List of Instruments to stream Prices for.
            snapshot:
                Flag that enables/disables the sending of a pricing snapshot
                when initially connecting to the stream.

        Returns:
            async_v20.interface.parser.Response containing the results from submitting the
            request
        """
        pass



from .base import Descriptor

__all__ = ['CandlestickGranularity', 'WeeklyAlignment']


class CandlestickGranularity(Descriptor):
    """The granularity of a candlestick
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'S5': '5 second candlesticks, minute alignment',
        'S10': '10 second candlesticks, minute alignment',
        'S15': '15 second candlesticks, minute alignment',
        'S30': '30 second candlesticks, minute alignment',
        'M1': '1 minute candlesticks, minute alignment',
        'M2': '2 minute candlesticks, hour alignment',
        'M4': '4 minute candlesticks, hour alignment',
        'M5': '5 minute candlesticks, hour alignment',
        'M10': '10 minute candlesticks, hour alignment',
        'M15': '15 minute candlesticks, hour alignment',
        'M30': '30 minute candlesticks, hour alignment',
        'H1': '1 hour candlesticks, hour alignment',
        'H2': '2 hour candlesticks, day alignment',
        'H3': '3 hour candlesticks, day alignment',
        'H4': '4 hour candlesticks, day alignment',
        'H6': '6 hour candlesticks, day alignment',
        'H8': '8 hour candlesticks, day alignment',
        'H12': '12 hour candlesticks, day alignment',
        'D': '1 day candlesticks, day alignment',
        'W': '1 week candlesticks, aligned to start of week',
        'M': '1 month candlesticks, aligned to first day of the month'
    }


class WeeklyAlignment(Descriptor):
    """The day of the week to use for candlestick granularities with weekly alignment.
    """

    # Type checking
    typ = str

    # Valid values
    values = {
        'Monday': 'Monday',
        'Tuesday': 'Tuesday',
        'Wednesday': 'Wednesday',
        'Thursday': 'Thursday',
        'Friday': 'Friday',
        'Saturday': 'Saturday',
        'Sunday': 'Sunday'
    }



import os

from setuptools import setup, find_packages
version = '1.1.5a4'

def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()

setup(name='async_v20',
      version=version,
      description="Asynchronous wrapper for OANDA's v20 REST API",
      long_description=read('README.rst'),
      author='James Peter Schinner',
      author_email='james.peter.schinner@gmail.com',
      url='https://github.com/jamespeterschinner/async_v20',
      license='MIT',
      packages=find_packages(),
      install_requires=['aiohttp>=2.2.5',
                        'ujson>=1.35',
                        'yarl>=0.12.0',
                        'inflection>=0.3.1'],
      classifiers=['Programming Language :: Python :: 3.6', 'Development Status :: 3 - Alpha',
                   'Framework :: AsyncIO',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Topic :: Internet :: WWW/HTTP :: Session'],
      keywords='algorithmic trading oanda v20 REST asyncio ',
      )



import asyncio
import ujson as json
from time import time

import aiohttp

from .definitions.types import AccountID


async def sleep(s=0.0):
    await asyncio.sleep(s)


async def wait(time_delta, previous_time):
    waited = time() - previous_time
    if waited <= time_delta:
        wait_time = time_delta - waited
        print('WAITING: ', wait_time)
        await sleep(wait_time)
    return True


async def request_limiter(self):
    yield
    time_delta = 1 / self.max_requests_per_second
    while True:
        previous_time = time()
        yield self.session.request
        await wait(time_delta, previous_time)

async def initializer(self):

    print('Initializing')

    self.initialized = 1

    await self.request.asend(None)

    conn = aiohttp.TCPConnector(limit=self.max_simultaneous_connections)

    headers = {'Content-Type': 'application/json', 'Connection': 'keep-alive',
               'OANDA-Agent': self.application}

    self.session = aiohttp.ClientSession(json_serialize=json.dumps, headers=headers, connector=conn)

    # Get the first account listed in in accounts
    response = await self.list_accounts()
    print(response)
    # Get the corresponding AccountID for the provided token
    self.default_parameters.update({AccountID: response['accounts'][0].id})

    # Get Account snapshot and last transaction id
    # last transaction is automatically updated when the
    # response is parsed
    response = await self.get_account_details()
    self.account = response['account']

    # Get the list of all available instruments for this account
    response = await self.account_instruments()
    self.account_instruments = response['instruments']
    self.initialized = 2
    print('initialize complete')

    while True:
        yield True


from functools import wraps
import inspect

from async_v20.definitions.helpers import flatten_dict

def order_dict(obj):
    if isinstance(obj, dict):
        return sorted((k, order_dict(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(order_dict(x) for x in obj)
    else:
        return obj



from collections import namedtuple

import pytest
from async_v20.definitions.types import Account, AccountSummary, AccountProperties
from async_v20.definitions.types import Price, Position
from async_v20.endpoints.account import GETAccountID, GETAccountIDSummary, GETAccounts
from async_v20.endpoints.annotations import LastTransactionID
from async_v20.endpoints.instrument import GETInstrumentsCandles
from async_v20.endpoints.pricing import GETPricingStream
from async_v20.interface.parser import _create_response
from async_v20.interface.parser import _rest_response
from async_v20.interface.parser import _stream_parser

from tests.data.json_data import GETAccountIDSummary_response
from tests.data.json_data import GETAccountID_response
from tests.data.json_data import GETAccounts_response
from tests.data.json_data import GETInstrumentsCandles_response
from tests.data.stream_data import price_bytes
from .helpers import order_dict


@pytest.fixture
@pytest.mark.asyncio
async def stream_generator():
    """FIXTURE to simulate data stream"""
    async def agen():
        while True:
            yield price_bytes # DATA
    stream = (i async for i in agen())
    yield stream
    del stream

@pytest.mark.asyncio
async def test_response_generator(stream_generator):
    """TEST stream_generator fixture works as intended"""
    result = list()
    result.append(await stream_generator.asend(None))
    result.append(await stream_generator.asend(None))
    assert result == [price_bytes, price_bytes]


@pytest.fixture()
def stream_response(stream_generator):
    """FIXTURE to simulate an `aiohttp` stream response"""
    class StreamResponse:
        content = stream_generator
        status = 200

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass
    response = StreamResponse()
    yield response
    del response


@pytest.mark.asyncio
async def test_stream_parser_creates_price_object(stream_response):
    async_gen = _stream_parser(stream_response, GETPricingStream)
    iteration = await async_gen.asend(None)
    assert type(iteration.get('PRICE')) == Price


@pytest.fixture()
def rest_response():
    """FIXTURE to simulate an `aiohttp` rest response"""
    class RestResponse(object):
        raw_headers = {}

        def __init__(self, data, status=200):
            self.data = data
            self.status = status

        async def json(self):
            return self.data

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass

    yield RestResponse
    del RestResponse


@pytest.fixture
def client_instance():
    """FIXTURE to simulate client object"""
    client = namedtuple('client', ['default_parameters'])
    yield client(dict())


@pytest.mark.asyncio
async def test_rest_response_builds_account_object_from_json_response(client_instance, rest_response):
    result = await _rest_response(client_instance, rest_response(GETAccountID_response), GETAccountID)
    # Ensure the result contains an 'account'
    assert 'account' in result
    # Ensure that 'account' is indeed an Account
    assert type(result['account']) == Account
    # Ensure that the Account has expected attributes
    assert hasattr(result['account'], 'data')
    # Ensure that account.positions is a list and contains a position
    # This tests that the Array object created the correct object
    assert type(result['account'].positions) == list
    assert type(result['account'].positions[0]) == Position

@pytest.mark.asyncio
async def test_rest_response_builds_account_summary_from_json_response(client_instance, rest_response):
    result = await _rest_response(client_instance, rest_response(GETAccountIDSummary_response), GETAccountIDSummary)
    # Ensure the result contains an 'account'
    assert 'account' in result
    # Ensure that 'account' is indeed an Account
    assert type(result['account']) == AccountSummary
    # Ensure that the Account has expected attributes
    assert hasattr(result['account'], 'data')

def test_array_object_creates_list_of_objects():
    array = GETAccounts.responses[200]['accounts']
    result = array(GETAccounts_response['accounts'])
    assert type(result) == list
    assert type(result[0]) == AccountProperties
    assert hasattr(result[0], 'id')


@pytest.mark.asyncio
async def test_rest_response_builds_array_account_properties(client_instance, rest_response):
    result = await _rest_response(client_instance, rest_response(GETAccounts_response), GETAccounts)
    # Ensure the result contains an 'account'
    print(result)
    assert 'accounts' in result
    # Ensure that 'account' is indeed an Account
    assert type(result['accounts']) == list
    assert type(result['accounts'][0]) ==  AccountProperties

@pytest.mark.asyncio
async def test_rest_response_updates_client_default_parameters(client_instance, rest_response):
    await _rest_response(client_instance, rest_response(GETAccountID_response), GETAccountID)
    # Ensure default_parameters is updated
    assert client_instance.default_parameters[LastTransactionID] == str(14)

@pytest.mark.asyncio
@pytest.mark.parametrize('json_body, schema', [(GETInstrumentsCandles_response, GETInstrumentsCandles.responses[200]),
                                               (GETAccounts_response, GETAccounts.responses[200]),
                                               (GETAccountIDSummary_response, GETAccountIDSummary.responses[200])])
async def test_conversion_from_server_json_to_response_object_to_json_equal(json_body, schema):
    response = await _create_response(json_body, schema)
    response_json = response.json_dict()
    pretty_json_body = order_dict(json_body)
    pretty_response_json = order_dict(response_json)
    print('SERVER JSON:\n', pretty_json_body)
    print('Response JSON:\n', pretty_response_json)
    assert response_json == json_body